# -*- coding: utf-8 -*-
"""
Spectral analysis, signal processing, and continuous (LFP/EEG) data preprocessing

This module is the base module for the `spectra` module. It contains all the
high-level API functions. Lower-level functions can be found in the other spectra modules.

Overview
--------
Functionality for computing frequency spectra as well as time-frequency (spectrogram) transforms.

Options to compute spectral analysis using multitaper, wavelet, band-pass filtering, or spectral
burst analysis methods.

Options to return full complex spectral data, spectral power, phase, real or imaginary part, etc.

Also includes functions for preprocessing, postprocessing, plotting of continuous/spectral data.

Most functions perform operations in a mass-univariate manner. This means that
rather than embedding function calls in for loops over channels, trials, etc., like this::

    for channel in channels:
        for trial in trials:
            results[trial,channel] = compute_something(data[trial,channel])

You can instead execute a single call on ALL the data, labeling the relevant axis
for the computation (usually time here), and it will run in parallel (vectorized)
across all channels, trials, etc. in the data, like this:

``results = compute_something(data, axis)``


Function list
-------------
General spectral analysis
^^^^^^^^^^^^^^^^^^^^^^^^^
- spectrum :          Frequency spectrum of data
- spectrogram :       Time-frequency spectrogram of data
- power_spectrum :    Power spectrum of data
- power_spectrogram : Power of time-frequency transform
- phase_spectrogram : Phase of time-frequency transform

Wavelet spectral analysis (spectra.wavelet)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- wavelet_spectrum :    Wavelet-based frequency spectrum
- wavelet_spectrogram : Time-frequency continuous wavelet transform
- compute_wavelets :    Compute wavelets for use in wavelet spectral analysis
- wavelet_bandwidth :   Compute time,frequency bandwidths for set of wavelets
- wavelet_edge_extent : Compute extent of edge effects for set of wavelets

Multitaper spectral analysis (spectra.multitaper)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- multitaper_spectrum :     Multitaper (DPSS) frequency spectrum
- multitaper_spectrogram :  Multitaper (DPSS) time-frequency spectrogram
- compute_tapers :          Compute DPSS tapers for use in multitaper spectral analysis

Bandpass-filtering spectral analysis (spectra.bandfilter)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- bandfilter_spectrum :     Band-filtered frequency spectrum
- bandfilter_spectrogram :  Band-filtered, Hilbert-transformed time-frequency of data
- set_filter_params :       Set filter coefficients for use in band-filtered analysis

Other spectral analyses
^^^^^^^^^^^^^^^^^^^^^^^
- itpc :                Intertrial phase clustering (analysis of phase locking to trial events)
- burst_analysis :      Compute oscillatory burst analysis of Lundqvist et al 2016

Preprocessing (spectra.preprocess)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- cut_trials :          Cut LFPs/continuous data into trial segments
- realign_data :        Realign LFPs/continuous data to new within-trial event
- remove_dc :           Remove constant DC component of signals
- remove_evoked :       Remove phase-locked evoked potentials from signals

Postprocesssing (spectra.postprocess)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- pool_freq_bands :         Average spectral data within set of frequency bands
- pool_time_epochs :        Average spectral data within set of time epochs
- one_over_f_norm :         Normalize to correct for 1/f distribution of spectral power

Utilities (spectra.utils)
^^^^^^^^^^^^^^^^^^^^^^^^^
- get_freq_sampling :       Frequency sampling vector for a given FFT-based computation
- complex_to_spec_type :    Convert complex Fourier transform output to power/phase/real/imag/etc.
- one_sided_to_two_sided :  Convert 1-sided Fourier transform output to 2-sided equivalent
- simulate_oscillation :    Generates simulated oscillation-in-noise data

Plotting
^^^^^^^^
- plot_spectrum :       Plot frequency spectrum as a line plot, handling freq axis properly
- plot_spectrogram :    Plot time-frequency spectrogram as a heatmap plot


Dependencies
------------
- pyfftw :              Python wrapper around FFTW, the speedy FFT library

Function reference
------------------
"""
# Created on Thu Oct  4 15:28:15 2018
#
# @author: sbrincat

from math import floor, ceil, log2, sqrt
import numpy as np
import matplotlib.pyplot as plt

from spynal.utils import set_random_seed, iarange
from spynal.spikes import _spike_data_type, times_to_bool
from spynal.plots import plot_line_with_error_fill, plot_heatmap
from spynal.spectra.wavelet import wavelet_spectrum, wavelet_spectrogram, compute_wavelets, \
                                   wavelet_bandwidth, wavelet_edge_extent
from spynal.spectra.multitaper import multitaper_spectrum, multitaper_spectrogram
from spynal.spectra.bandfilter import bandfilter_spectrum, bandfilter_spectrogram
from spynal.spectra.postprocess import one_over_f_norm, pool_freq_bands 
from spynal.spectra.helpers import _infer_freq_scale, _frequency_plot_settings


# =============================================================================
# General spectral analysis functions
# =============================================================================
def spectrum(data, smp_rate, axis=0, method='multitaper', data_type='lfp', spec_type='complex',
             removeDC=True, **kwargs):
    """
    Compute frequency spectrum of data using given method

    Parameters
    ----------
    data : ndarray,shape=(...,n_samples,...)
        Data to compute spectral analysis of.
        Arbitrary shape; spectral analysis is computed along `axis`.

    smp_rate : scalar
        Data sampling rate (Hz)

    axis : int, default: 0 (1st axis)
        Axis of `data` to do spectral analysis on (usually time dimension).

    method : {'multitaper','wavelet','bandfilter'}, default: 'multitaper'
        Specific spectral analysis method to use:

        - 'multitaper' : Multitaper spectral analysis in :func:`.multitaper_spectrum`
        - 'wavelet' : Wavelet analysis in :func:`.wavelet_spectrum`
        - 'bandfilter' : Bandpass filtering in :func:`.bandfilter_spectrum`

    data_type : {'lfp','spike'}, default: 'lfp'
        Type of signal in data

    spec_type : {'complex','power','phase','real','imag'}, default: 'complex'
        Type of spectral signal to return. See :func:`.complex_to_spec_type` for details.

    removeDC : bool, default: True
        If True, subtracts off mean DC component across `axis`, making signals zero-mean
        before spectral analysis.

    **kwargs :
        All other kwargs passed directly to method-specific spectrum function

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs,...), dtype=complex or float.
        Frequency spectrum of given type computed with given method.
        Frequency axis is always inserted in place of `axis`.
        Note: 'multitaper' method will return with additional taper
        axis inserted after just after `axis` if `keep_tapers` is True.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqs,) or (n_freqbands,2)
        For `method` == 'bandfilter': List of (low,high) cut frequencies (Hz) used to
        generate `spec`, shape=(n_freqbands,2)
        For other methods: List of frequencies in `spec` (Hz), shape=(n_freqs,)
    """
    method = method.lower()
    assert data_type in ['lfp','spike'], \
        ValueError("<data_type> must be 'lfp' or 'spike' ('%s' given)" % data_type)

    if method == 'multitaper':      spec_fun = multitaper_spectrum
    elif method == 'wavelet':       spec_fun = wavelet_spectrum
    elif method == 'bandfilter':    spec_fun = bandfilter_spectrum
    else:
        raise ValueError("Unsupported value set for <method>: '%s'" % method)

    spec,freqs = spec_fun(data,smp_rate,axis=axis,data_type=data_type,spec_type=spec_type,
                          removeDC=removeDC, **kwargs)

    return spec, freqs


def spectrogram(data, smp_rate, axis=0, method='wavelet', data_type='lfp', spec_type='complex',
                removeDC=True, **kwargs):
    """
    Compute time-frequency transform of data using given method

    Parameters
    ----------
    data : ndarray, shape=(...,n_samples,...)
        Data to compute spectral analysis of.
        Arbitrary shape; spectral analysis is computed along `axis`.

    smp_rate : scalar
        Data sampling rate (Hz)

    axis : int, default: 0 (1st axis)
        Axis of `data` to do spectral analysis on (usually time dimension).

    method : {'multitaper','wavelet','bandfilter','burst'}, default: 'wavelet'
        Specific spectral analysis method to use:

        - 'multitaper' : Multitaper spectral analysis in :func:`.multitaper_spectrogram`
        - 'wavelet' : Wavelet analysis in :func:`.wavelet_spectrogram`
        - 'bandfilter' : Bandpass filtering in :func:`.bandfilter_spectrogram`
        - 'burst' : Oscillatory burst analysis in :func:`.burst_analysis`

    data_type : {'lfp','spike'}, default: 'lfp'
        Type of signal in data

    spec_type : {'complex','power','phase','real','imag'}, default: 'complex'
        Type of spectral signal to return. See :func:`.complex_to_spec_type` for details.

    removeDC : bool, default: True
        If True, subtracts off mean DC component across `axis`, making signals zero-mean
        before spectral analysis.

    **kwargs :
        All other kwargs passed directly to method-specific spectrogram function

    Returns
    -------
    spec : ndarray, shape=(...,n_freqs,n_timepts,...), dtype=complex or float.
        Time-frequency spectrogram of given type computed with given method.
        Frequency axis is always inserted just before time axis.
        Note: 'multitaper' method will return with additional taper
        axis inserted between freq and time axes if `keep_tapers` is True.
        dtype is complex if `spec_type` is 'complex', float otherwise.

    freqs : ndarray, shape=(n_freqs,) or (n_freqbands,2)
        For `method` == 'bandfilter': List of (low,high) cut frequencies (Hz) used to
        generate `spec`, shape=(n_freqbands,2)
        For other methods: List of frequencies in `spec` (Hz), shape=(n_freqs,)

    timepts : ndarray, shape=(n_timepts,)
        List of time points / time window centers in `spec` (in s, referenced to start of data)
    """
    method = method.lower()

    # Special case: Lundqvist oscillatory burst analysis
    if (spec_type == 'burst') or (method == 'burst'):
        assert data_type == 'lfp', ValueError("<data_type> must be 'lfp' for burst analysis")

        spec,freqs,timepts = burst_analysis(data, smp_rate, axis=axis, removeDC=removeDC, **kwargs)

    else:
        assert data_type in ['lfp','spike'], \
            ValueError("<data_type> must be 'lfp' or 'spike' ('%s' given)" % data_type)

        if method == 'wavelet':         spec_fun = wavelet_spectrogram
        elif method == 'multitaper':    spec_fun = multitaper_spectrogram
        elif method == 'bandfilter':    spec_fun = bandfilter_spectrogram
        else:
            raise ValueError("Unsupported value set for <method>: '%s'" % method)

        spec,freqs,timepts = spec_fun(data, smp_rate, axis=axis, data_type=data_type,
                                      spec_type=spec_type, removeDC=removeDC, **kwargs)

    return spec, freqs, timepts


def power_spectrum(data, smp_rate, axis=0, method='multitaper', **kwargs):
    """
    Convenience wrapper around spectrum() to compute **power** spectrum of data with given method

    See :func:`spectrum` for details
    """
    return spectrum(data, smp_rate, axis=axis, method=method, spec_type='power', **kwargs)


def power_spectrogram(data, smp_rate, axis=0, method='wavelet', **kwargs):
    """
    Convenience wrapper around spectrogram() to compute time-frequency **power** with given method

    See :func:`spectrogram` for details
    """
    return spectrogram(data, smp_rate, axis=axis, method=method, spec_type='power', **kwargs)


def phase_spectrogram(data, smp_rate, axis=0, method='wavelet', **kwargs):
    """
    Convenience wrapper around spectrogram() to compute **phase** of time-frequency transform

    See :func:`spectrogram` for details
    """
    return spectrogram(data, smp_rate, axis=axis, method=method, spec_type='phase', **kwargs)


# =============================================================================
# Other spectral analysis functions
# =============================================================================
def itpc(data, smp_rate, axis=0, method='wavelet', itpc_method='PLV', trial_axis=None, **kwargs):
    """
    Intertrial phase clustering (ITPC) measures frequency-specific phase locking of continuous
    neural activity (eg LFPs) to trial events. A spectral analog (roughly) of evoked potentials.

    Complex time-frequency representation is first computed (using some method), then the
    complex vector mean is computed across trials, and it's magnitude is returned as ITPC.

    aka "intertrial coherence", "intertrial phase-locking value/factor"

    Parameters
    ----------
    data : ndarray, shape=(...,n_timepts,...,n_trials,...)
        Data to compute ITPC of, aligned to some within-trial event.
        Arbitrary shape; spectral analysis is computed along `axis` (usually time),
        and ITPC is computed along `trial_axis`.

    smp_rate : scalar
        Data sampling rate (Hz)

    axis : int, default: 0 (1st axis)
        Axis of `data` to do spectral analysis on (usually time dimension).

    method : {'multitaper','wavelet','bandfilter'}, default: 'wavelet'
        Specific underlying spectral analysis method to use:

        - 'multitaper' : Multitaper spectral analysis in :func:`.multitaper_spectrogram`
        - 'wavelet' : Wavelet analysis in :func:`.wavelet_spectrogram`
        - 'bandfilter' : Bandpass filtering in :func:`.bandfilter_spectrogram`

    itpc_method : {'PLV','Z','PPC'}, default: 'PLV'
        Method to use for computing intertrial phase clustering:

        - 'PLV' : Phase locking value (length of cross-trial complex vector mean)
            Standard/traditional measure of ITPC, but is positively biased
            for small n, and is biased > 0 even for no phase clustering.
        - 'Z' :   Rayleigh's Z normalization of PLV (Z = n*PLV**2). Reduces small-n bias.
        - 'PPC' : Pairwise Phase Consistency normalization (PPC = (n*PLV**2 - 1)/(n - 1))
            of PLV. Debiased and has expected value 0 for no clustering.

    **kwargs :
        All other kwargs passed directly to method-specific spectrogram function

    Returns
    -------
    ITPC : ndarray, shape=(...,n_freqs,n_timepts,...)
        Time-frequency spectrogram representation of intertrial phase clustering.
        Frequency axis is always inserted just before time axis, and trial axis is
        removed, but otherwise shape is same as input `data`.

    freqs : ndarray, shape=(n_freqs,) or (n_freqbands,2)
        For `method` == 'bandfilter': List of (low,high) cut frequencies (Hz) used to
        generate `ITPC`, shape=(n_freqbands,2)
        For other methods: List of frequencies in `spec` (Hz), shape=(n_freqs,)

    timepts : ndarray, shape=(n_timepts,)
        List of time points / time window centers in `ITPC` (in s, referenced to start of data)

    References
    ----------
    Cohen "Analyzing Neural Time Series Data" http://dx.doi.org/10.7551/mitpress/9609.001.0001 Ch. 19
    """
    method = method.lower()
    itpc_method = itpc_method.lower()
    if axis < 0:        axis = data.ndim + axis
    if trial_axis < 0:  trial_axis = data.ndim + trial_axis

    n = data.shape[trial_axis]

    if method == 'wavelet':         spec_fun = wavelet_spectrogram
    elif method == 'multitaper':    spec_fun = multitaper_spectrogram
    elif method == 'bandfilter':    spec_fun = bandfilter_spectrogram
    else:
        raise ValueError("Unsupported value set for <method>: '%s'" % method)

    if method == 'multitaper': kwargs.update(keep_tapers=True)

    # Compute spectrogram using given method
    spec,freqs,timepts = spec_fun(data, smp_rate, axis=axis, spec_type='complex', **kwargs)
    # Account for new frequency (and/or taper) axis prepended before time axis
    n_new_axes = 2 if method == 'multitaper' else 1
    if trial_axis >= axis: trial_axis += n_new_axes
    axis += n_new_axes
    if method == 'multitaper': taper_axis = axis - 1

    spec = spec / np.abs(spec)     # Normalize spectrum to unit length

    # Compute mean resultant length (aka PLV) = length of complex vector mean across trials
    # For multitaper spectral analysis, take mean across both trials and tapers
    reduce_axes = (trial_axis,taper_axis) if method == 'multitaper' else trial_axis
    ITPC = np.abs(spec.mean(axis=reduce_axes))

    # PLV -- we are done, return ITPC as-is
    if itpc_method in ['plv','plf']:         pass

    # PPC -- debiasing normalization
    elif itpc_method in ['ppc']:             ITPC = (n*ITPC**2 - 1) / (n - 1)

    # Rayleigh's Z -- debiasing normalization
    elif itpc_method in ['z','rayleighz']:   ITPC = n*ITPC**2

    else:
        raise ValueError("%s in an unsupported ITPC method" % itpc_method)

    return ITPC, freqs, timepts

intertrial_phase_clustering = itpc
""" Alias of :func:`itpc`. See there for details. """


def burst_analysis(data, smp_rate, axis=0, trial_axis=-1, threshold=2, min_cycles=3,
                   method='wavelet', spec_type='power', freq_exp=None,
                   bands=((20,35),(40,65),(55,90),(70,100)),
                   window=None, timepts=None, **kwargs):
    """
    Oscillatory burst analysis of Lundqvist et al 2016.

    Computes oscillatory power, z-scores within each frequency band, thresholds at given
    z `threshold`, labels as burst "ON" times timepoints > `threshold` for at least
    `min_cycles` duration.

    To compute burst rate, simply take mean of computed `bursts` across trial axis.

    Default argument values approximate analysis as implemented in Lundqvist 2016.

    Parameters
    ----------
    data : ndarray, shape=(...,n_samples,...)
        Data to compute spectral analysis of.
        Arbitrary shape; spectral analysis is computed along `axis`.

    smp_rate : scalar
        Data sampling rate (Hz)

    smp_rate : scalar
        Data sampling rate (Hz)

    axis : int, default: 0 (1st axis)
        Axis of `data` to do spectral analysis on (usually time dimension).

    trial_axis : int, default: -1 (last axis of `data`)
        Axis of `data` corresponding to trials/observations.

    threshold : scalar, default: 2 (2 SDs above mean)
        Threshold power level for detecting bursts, given in SDs above the mean

    min_cycles : scalar, default: 3
        Minimal length of contiguous above-threshold period to be counted as a burst,
        given in number of oscillatory cycles at each frequency (or band center).

    method : {'multitaper','wavelet','bandfilter'}, default: 'wavelet'
        Specific spectral analysis method to use:

        - 'multitaper' : Multitaper spectral analysis in :func:`.multitaper_spectrogram`
        - 'wavelet' : Wavelet analysis in :func:`.wavelet_spectrogram`
        - 'bandfilter' : Bandpass filtering in :func:`.bandfilter_spectrogram`

        Note: In the original paper, multitaper was used, but all three
        methods were claimed to produced similar results.

    spec_type : {'power','magnitude'}, default: 'power'
        Type of spectral signal to compute:

        - 'power' : Spectral power, ie square of signal envelope
        - 'magnitude' : Square root of power, ie signal envelope

    freq_exp : float, default: None (do no frequency normalization)
        This can be used to normalize out 1/f^a effects in power before
        band-pooling and burst detection). This gives the exponent on the frequency
        ('a' in 1/f^a).  Set = 1 to norm by 1/f.  Set = None for no normalization.

    bands : array-like, shape=(n_freqbands,2), default: ((20,35),(40,65),(55,90),(70,100))
        List of (low,high) cut frequencies for each band to compute bursts within.
        Set low cut = 0 for low-pass, set high cut >= smp_rate/2 for high-pass,
        otherwise assumes band-pass. Default samples ~ beta, low/med/high gamma.
        Set = None to compute bursts at each frequency in spectral transform.

    window : array-like, shape=(2,), default: None (compute over entire data time range)
        Optional (start,end) of time window to compute mean,SD for burst amplitude threshold
        within (in same units as `timepts`).

    timepts : array-like, shape=(n_timepts,), default: None
        Time sampling vector for data (usually in s).
        Necessary if `window` is set, but unused otherwise.

    **kwargs :
        Any other kwargs passed directly to :func:`spectrogram` function

    Returns
    -------
    bursts : ndarray, shape=(...,n_freq[band]s,n_timepts_out,...), dtype=bool
        Binary array labelling timepoints within detected bursts in each trial and frequency band.
        Same shape as `data`, but with frequency axis prepended immediately before time `axis`.

    freqs : ndarray, shape=(n_freq[band],)
        List of center frequencies of bands in `bursts`

    timepts : ndarray=(n_timepts_out,)
        List of time points / time window centers in `bursts` (in s, referenced to start of data).

    References
    ----------
    Lundqvist et al (2016) Neuron https://doi.org/10.1016/j.neuron.2016.02.028
    Lundqvist et al (2018) Nature Comms https://doi.org/10.1038/s41467-017-02791-8
    """
    # todo  Gaussian fits for definining burst f,t extent?
    # todo  Option input of spectral data?
    # todo  Add optional sliding trial window for mean,SD
    method = method.lower()
    spec_type = spec_type.lower()
    if bands is not None: bands = np.asarray(bands)
    if bands.ndim == 1: bands = bands[np.newaxis,:] # Ensure bands is (n_bands,2) even if n_bands=1
    if axis < 0: axis = data.ndim + axis
    if (trial_axis is not None) and (trial_axis < 0):  trial_axis = data.ndim + trial_axis

    assert axis != trial_axis, \
        ValueError("Time and trial axes can't be same. Set trial_axis=None if no trials in data.")
    if window is not None:
        assert len(window) == 2, \
            ValueError("Window for computing mean,SD should be given as (start,end) (len=2)")
        assert timepts is not None, \
            ValueError("Need to input <timepts> to set a window for computing mean,SD")
    assert spec_type in ['power','magnitude'], \
        ValueError("spec_type must be 'power'|'magnitude' for burst analysis (%s given)"
                   % spec_type)

    # Move array axes so time axis is 1st and trials last (n_timepts,...,n_trials)
    if (axis == data.ndim-1) and (trial_axis == 0):
        data = np.swapaxes(data, axis, trial_axis)
    else:
        if axis != 0:
            data = np.moveaxis(data,axis,0)
        # If data has no trial axis, temporarily append singleton to end to simplify code
        if trial_axis is None:
            data = data[...,np.newaxis]
        elif trial_axis != data.ndim-1:
            data = np.moveaxis(data, trial_axis, -1)
    data_shape = data.shape
    data_ndim = data.ndim
    # Standardize data array to shape (n_timepts,n_data_series,n_trials)
    if data_ndim > 3:       data = data.reshape((data.shape[0],-1,data.shape[-1]))
    elif data_ndim == 2:    data = data[:,np.newaxis,:]
    n_timepts,n_series,n_trials = data.shape

    # Set default values to appoximate Lundqvist 2016 analysis, unless overridden by inputs
    if method == 'wavelet':
        # Sample frequency at 1 Hz intervals from min to max frequency in requested bands
        if ('freqs' not in kwargs) and (bands is not None):
            kwargs['freqs'] = np.arange(bands.min(),bands.max()+1,1)

    # For bandfilter method, if frequency bands not set explicitly, set it with value for <bands>
    elif method == 'bandfilter':
        if ('freqs' not in kwargs) and (bands is not None): kwargs['freqs'] = bands

    # Compute time-frequency power from raw data -> (n_freqs,n_timepts,n_data_series,n_trials)
    data,freqs,times = spectrogram(data, smp_rate, axis=0, method=method, spec_type=spec_type,
                                   **kwargs)
    timepts = times + timepts[0] if timepts is not None else times
    n_timepts = len(times)
    dt = np.mean(np.diff(times))

    # Normalize computed power by 1/f**exp to normalize out 1/f distribution of power
    if freq_exp is not None:
        data = one_over_f_norm(data, axis=0, freqs=freqs, exponent=freq_exp)

    # If requested, pool data within given frequency bands
    # (skip for bandfilter spectral analysis, which already returns frequency bands)
    if (method != 'bandfilter') and (bands is not None):
        data = pool_freq_bands(data, bands, axis=0, freqs=freqs, func='mean')
        freqs = bands
    n_freqs = data.shape[0]

    # Compute mean,SD of each frequency band and data series (eg channel)
    # across all trials (axis -1) and timepoints (axis 1)
    if window is None:
        mean = data.mean(axis=(1,-1), keepdims=True)
        sd   = data.std(axis=(1,-1), ddof=0, keepdims=True)

    # Compute mean,SD of each freq band/channel across all trials and timepoints w/in time window
    else:
        tbool = (timepts >= window[0]) & (timepts <= window[1])
        mean = data.compress(tbool,axis=1).mean(axis=(1,-1), keepdims=True)
        sd   = data.compress(tbool,axis=1).std(axis=(1,-1), ddof=0, keepdims=True)

    # Compute z-score of data and threshold -> boolean array of candidate burst times
    z = (data - mean) / sd
    bursts = z > threshold

    tsmps = np.arange(n_timepts)


    def _screen_bursts(burst_bool, min_samples, start):
        """ Subfunction to evaluate/detect bursts in boolean time series of candidate bursts """
        # Find first candidate burst in trial (timepoints of all candidate burst times)
        if start is None:   on_times = np.nonzero(burst_bool)[0]
        # Find next candidate burst in trial
        else:               on_times = np.nonzero(burst_bool & (tsmps > start))[0]

        # If no (more) bursts in time series, return data as is, we are done
        if len(on_times) == 0:  return burst_bool
        # Otherwise, get onset index of first/next candidate burst
        else:                   onset = on_times[0]

        # Find non-burst timepoints in remainder of time series
        off_times = np.nonzero(~burst_bool & (tsmps > onset))[0]

        # If no offset found, burst must extend to end of data
        if len(off_times) == 0: offset = len(burst_bool)
        # Otherwise, get index of offset of current burst = next off time - 1
        else:                   offset = off_times[0] - 1

        # Determine if length of current candidate burst meets minimum duration
        # If not, delete it from data (set all timepoints w/in it to False)
        burst_len = offset - onset + 1
        if burst_len < min_samples:  burst_bool[onset:(offset+1)] = False

        # todo trim bursts to half-max point? (using Gaussian fits or raw power?)

        # If offset is less than minimum burst length from end of data, we are done.
        # Ensure no further timepoints are labelled "burst on" and return data
        if (len(burst_bool) - offset) < min_samples:
            burst_bool[(offset+1):] = False
            return burst_bool
        # Otherwise, call function recursively, now starting search just after current burst offset
        else:
            return _screen_bursts(burst_bool,min_samples,offset+1)


    # Screen all candidate bursts across freqs/trials/chnls to ensure they meet minimum duration
    for i_freq,freq in enumerate(freqs):
        # Compute center frequency of frequency band
        if not np.isscalar(freq): freq = np.mean(freq)
        # Convert minimum length in oscillatory cycles -> samples
        min_samples = ceil(min_cycles * (1/freq) / dt)

        for i_trial in range(n_trials):
            for i_series in range(n_series):
                # Extract burst time series for current (frequency, data series, trial)
                series = bursts[i_freq,:,i_series,i_trial]
                if not series.any(): continue

                bursts[i_freq,:,i_series,i_trial] = _screen_bursts(series,min_samples,None)

    # Reshape data array to ~ original dimensionality -> (n_freqs,n_timepts,...,n_trials)
    if data_ndim > 3:       bursts = bursts.reshape((n_freqs,n_timepts,data_shape[1:-1],n_trials))
    elif data_ndim == 2:    bursts = bursts.squeeze(axis=2)

    # Move array axes back to original locations
    if (axis == data_ndim-1) and (trial_axis == 0):
        bursts = np.moveaxis(bursts,-1,0)   # Move trial axis back to 0
        bursts = np.moveaxis(bursts,1,-1)   # Move freq axis to end
        bursts = np.moveaxis(bursts,1,-1)   # Move time axis to end (after freq)
    else:
        if axis != 0:                   bursts = np.moveaxis(bursts,(0,1),(axis,axis+1))
        if trial_axis is None:          bursts = np.squeeze(bursts,-1)
        elif trial_axis != data_ndim-1:
            if trial_axis > axis:       bursts = np.moveaxis(bursts,-1,trial_axis+1)
            else:                       bursts = np.moveaxis(bursts,-1,trial_axis)

    return bursts, freqs, timepts


# =============================================================================
# Plotting functions
# =============================================================================
def plot_spectrum(freqs, data, ax=None, ylim=None, color=None, **kwargs):
    """
    Plot frequency spectrum as a line plot.

    Parameters
    ----------
    freqs : array-like, shape=(n_freqs,)
        Frequency sampling (x-axis) vector for data (Hz).
        May be linearly or logarithmically sampled; we handle appropriately.

    data : ndarray, shape=(n_freqs,)
        Frequency spectrum data to plot (y-axis)

    ax : Pyplot Axis object, default: plt.gca()
        Axis to plot into

    ylim : array-like, shape=(2,), Default: (data.min(),data.max()) +/- 5%
        Plot y-axis limits: (min,max)

    color : Color spec, default: <Matplotlib default plot color>
        Color to plot in

    **kwargs :
        Any additional keyword args are interpreted as parameters of plt.axes()
        (settable Axes object attributes) or plt.plot() (Line2D object attributes),
        and passsed to the proper function.

    Returns
    -------
    lines : List of Line2D objects
        Output of plt.plot()

    ax : Axis object
        Axis plotted into
    """
    freqs,fticks,fticklabels = _frequency_plot_settings(freqs)

    lines, _, ax = plot_line_with_error_fill(freqs, data, ax=ax, ylim=ylim, color=color, **kwargs)

    plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
    ax.set_xticks(fticks)
    ax.set_xticklabels(fticklabels)
    # No need to return list of lists if only plotting one data series
    if (data.ndim == 1) or (data.shape[0] == 1): lines = lines[0]

    return lines, ax


def plot_spectrogram(timepts, freqs, data, ax=None, clim=None, cmap='viridis', **kwargs):
    """
    Plot time-frequency spectrogram as a heatmap plot.

    Parameters
    ----------
    timepts : array-like, shape=(n_timepts,)
        Time sampling (x-axis) vector for data

    freqs : array-like, shape=(n_freqs,)
        Frequency sampling (y-axis) vector for data (Hz).
        May be linearly or logarithmically sampled; we handle appropriately.

    data : ndarray, shape=(n_freqs,n_timepts)
        Time-frequency (spectrogam) data to plot on color axis

    ax : Pyplot Axis object, default: plt.gca()
        Axis to plot into

    clim : array-like, shape=(2,), default: (data.min(),data.max())
        Color axis limits: (min,max)

    cmap  : str | Colormap object. default: 'viridis' (linear dark-blue to yellow colormap)
        Colormap to plot heatmap in, given either as name of matplotlib colormap or custom
        matplotlib.colors.Colormap object instance.

    **kwargs :
        Any additional keyword args are interpreted as parameters of :func:`plt.axes`
        (settable Axes object attributes) or :func:`plt.imshow` (AxesImage object attributes).

    Returns
    -------
    img : AxesImage object
        Output of ax.imshow(). Allows access to image properties.

    ax : Axis object
        Axis plotted into.
    """
    freqs,fticks,fticklabels = _frequency_plot_settings(freqs)

    img, ax = plot_heatmap(timepts, freqs, data, ax=ax, clim=clim, cmap=cmap,
                           origin='lower', **kwargs)

    plt.grid(axis='y',color=[0.75,0.75,0.75],linestyle=':')
    ax.set_yticks(fticks)
    ax.set_yticklabels(fticklabels)

    return img, ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Helper functions for matIO package """
from types import SimpleNamespace
from warnings import warn
import numpy as np
import pandas as pd

import scipy.io

# We check for xarray variables in _variables_to_mat, but don't require as dependency
try:
    import xarray as xr
    _HAS_XARRAY = True
except ImportError:
    _HAS_XARRAY = False
    # print("xarray module is not installed. No support for loading/saving xarray format")

from spynal.helpers import _isbinary, _enclose_in_object_array
# from spynal.matIO.matIO_7 import _v7_matlab_type, _process_v7_object

# Set=True to print out detailed debugging statements in loading functions
DEBUG = False


# =============================================================================
# General helper functions
# =============================================================================
def _structuredarray_to_dict(sarray, typemap=None, extract_item=True, extract_items=None,
                             transpose=False, level=0):
    """
    Convert Numpy structured array (which Matlab structs are initially
    imported as, using scipy.io.loadmat) to a dict, with type names as keys

    Parameters
    ----------
    sarray :  Numpy structured array

    Returns
    -------
    dic : dict, {sarray.dtype.names : sarray[key]}
    """
    # Convert structured array to dict
    dic = {name:sarray[name] for name in sarray.dtype.names}

    for name in dic.keys():
        if DEBUG: print('\t'*level, "'%s'" % name)
        # Infer Matlab variable type of struct field
        matlab_vbl_type = _v7_matlab_type(dic[name])

        # If specific field name is listed in <typemap>/<extract_items>, use given value
        python_vbl_type = typemap[name] if name in typemap else None
        extract_item = extract_items[name] if name in extract_items else extract_item

        dic[name] = _process_v7_object(dic[name], matlab_vbl_type=matlab_vbl_type,
                                       python_vbl_type=python_vbl_type, extract_item=extract_item,
                                       typemap=typemap, extract_items=extract_items,
                                       transpose=transpose, level=level)

    return dic


def _structuredarray_to_dataframe(sarray, typemap=None, extract_item=None, extract_items=None,
                                  transpose=False, level=0):
    """
    Convert Numpy structured array (which Matlab structs are initially imported
    as, using scipy.io.loadmat) to a Pandas DataFrame, with type names as keys.
    Intended for table-like Matlab scalar structs, which naturally fit into a
    DataFrame (though Matlab table variables are not currently loadable in Python).

    Parameters
    ----------
    sarray : Numpy structured array

    Returns
    -------
    df : Pandas DataFrame
        Columns are names and values of all the sarray.dtype.names
        whose associated values are vector-valued with length = modal length
        of all names. Rows are each entry from these vectors

    References
    ----------
    http://poquitopicante.blogspot.com/2014/05/loading-matlab-mat-file-into-pandas.html
    """
    # Convert structured array to dict {keys = sarray.dtype.names : values = sarray[key]}
    dic = _structuredarray_to_dict(sarray, typemap=typemap, extract_item=extract_item,
                                   extract_items=extract_items, transpose=transpose, level=level)

    # Convert dict to a Pandas DataFrame
    df = _dict_to_dataframe(dic)

    return df


def _get_matfile_version(filename):
    """
    Determines version of given MAT file (7.3 or 7/older) by reading its header

    Parameters
    ----------
    filename : str
        Full-path name of MAT-file to load from

    Returns
    -------
    version : float
        Version of MAT file: 7 (or 4,5 = older) | 7.3 (newer/HDF5)
    """
    # Read in first 10 bytes of MAT file header and convert to string ("with" auto-closes file)
    with open(filename, 'rb') as file:
        hdr = file.read(10).decode("utf-8")

    # Check for version 7 MAT file header (no clue why it says "5.0", but this is really the case)
    if 'MATLAB 5.0' in hdr:
        version = 7.0

    # Check for version 7.3 MAT file header
    elif 'MATLAB 7.3' in hdr:
        version = 7.3

    else:
        version = None
        print('WARNING: Unable to determine version of MAT file ''%s''' % filename)

    return version


def _extract_strings(x):
    """
    Deal with weird way Matlab cell-array-of-strings data structures are loaded.

    Convert them from Numpy array of objects, which are each trivial
    1-length arrays of strings, to just an array of strings themselves.
    """
    return np.asarray([(x[j][0] if len(x[j]) > 0 else '') for j in range(len(x))])


def _is_structured_array(array):
    """ Return True if input array is a Numpy structured array, False otherwise """
    # For v7.3 mat files loaded with h5py, structured arrays are ndarrays with dtype np.void
    # For v7 mat files loaded with scipy.io, structured arrays are an "np.void" type
    # Test for either version
    return (isinstance(array,np.ndarray) and (array.dtype.type is np.void)) \
         or isinstance(array,np.void)


def _dict_to_dataframe(dic):
    """
    Convert dict (which Matlab structs are initially imported as, using h5py)
    to a Pandas DataFrame, with type names as keys. Intended for table-like
    Matlab scalar structs, which naturally fit into a DataFrame
    (though Matlab tables themselves are not currently loadable in Python AKAIK).

    Parameters
    ----------
    dic : dict, {str : array-like, shape=(n_rows,)}
        Dictionary with keys to be used as column names and values that are same-length
        vectors. Values not fitting this criterion are not retained in output.

    Returns
    -------
    df : Pandas DataFrame, shape=(n_rows,n_columns)
        Columns are dict keys, for all associated values are vector-valued with
        length = modal length of all keys. Rows are each entry from these vectors.

    References
    ----------
    http://poquitopicante.blogspot.com/2014/05/loading-matlab-mat-file-into-pandas.html
    """
    # Special case: Matlab structs converted from Matlab tables may retain the
    # 'Properties' metadata attribute as a struct field. If so, we extract and
    #  remove it, convert it to dict, and below tack it onto df as an attribute.
    if 'Properties' in dic:
        metadata = dic.pop('Properties')

        # If metadata is a structured array, convert it to dict
        if _is_structured_array(metadata):
            metadata = _structuredarray_to_dict(metadata, typemap={'struct':'dict'},
                                                extract_item=True, transpose=False)
    else:
        metadata = None

    # Convert any scalar dict values -> ndarrays, and ensure ndarrays are
    # at least 1D (0-dimensional arrays mess up DF conversion and are weird)
    for key in dic.keys():
        if not isinstance(dic[key],np.ndarray) or (dic[key].ndim == 0):
            dic[key] = np.atleast_1d(dic[key])

    # Find length of each value in dict <dic> (to become DataFrame columns)
    lengths = [value.shape[0] if value.ndim > 0 else 1 for value in dic.values()]
    # Find modal value of column lengths
    # From: https://stackoverflow.com/questions/10797819/finding-the-mode-of-a-list
    height = max(set(lengths), key=lengths.count)

    # Reshape any (n,k>1) ndarrays -> (n,) vectors of k-tuples
    for key in dic.keys():
        if (dic[key].ndim > 1) and \
           (dic[key].shape[0] == height) and (dic[key].shape[1] > 1):
            dic[key] = _array_to_tuple_vector(dic[key])

    # Special case: height=1 DF with some tuple-valued columns--convert to (1,) object arrays
    if height == 1:
        for key in dic.keys():
            if dic[key].shape[0] > height: dic[key] = _enclose_in_object_array(dic[key])
            # HACK Convert empty (0-length) arrays to (1,) array with value = nan
            elif dic[key].shape[0] == 0:    dic[key] = np.atleast_1d(np.nan)

    columns = list(dic.keys())

    # Create a DataFrame with length-consistent key/value pairs as columns
    df = pd.DataFrame(dic,columns=columns,index=range(height))

    # Tack Matlab table-derived 'Properties' metadata onto df as attribute.
    if metadata is not None:
        # Setting a DF attribute with a dict throws an error.
        # Workaround here from https://stackoverflow.com/a/54137536
        df.metadata = SimpleNamespace()
        df.metadata = metadata
        # If 'Properties' had a 'RowNames' field, use that as DataFrame row index.
        if ('RowNames' in metadata):
            metadata['RowNames'] = np.atleast_1d(metadata['RowNames'])
            if metadata['RowNames'].ndim > 1: metadata['RowNames'] = metadata['RowNames'].squeeze()
            # Ensure RowNames was not empty, has expected length
            if not np.array_equal(metadata['RowNames'],[0,0]) and \
                (len(metadata['RowNames']) == df.shape[0]):
                df.index = metadata['RowNames']

    return df


def _array_to_tuple_vector(array):
    """
    Convert ndarray vector of shape (n_rows,n_cols) to (n_rows,) vector of
    n_cols length tuples, which can more easily be assembled into a DataFrame

    Parameters
    ----------
    array : ndarray, shape=(n_rows,n_cols)

    Returns
    -------
    tuples : ndarray, shape=(n_rows,) of tuples, shape=(n_cols,)
    """
    n = array.shape[0]
    # Create vector of size (n,) of same type as <array>
    tuples = np.empty((n,), dtype=object)
    for row in range(n):
        tuples[row] = tuple(array[row,:])

    return tuples


def _xarray_to_array(array):
    """
    Convert xarray DataArray -> Numpy array + dict with metadata attributes

    Parameters
    ----------
    array : xarray DataArray, shape=Any

    Returns
    -------
    array : ndarray, shape=Any
        Data values from DataArray `array`. Same shape and dtype as `array`.

    attrs : dict, {str : *}
        Metatdata attributes extracted from DataArray:

        - 'dims' :  tuple, shape=(n_dims,), of str. Dimension names.
        - 'coords' : tuple, shape=(n_dims,), of [len(dim),]. Coordinate indexes for each dimension.
        - 'name' :  str. Name of variable ('' if none set).
        - 'attrs':  OrderedDict. Any additional metadata ([] if none set).
    """
    # Extract metadata from variable into dict
    attrs = {'dims':   array.dims,
             'coords': tuple([array.coords[dim].values for dim in array.dims]),
             'name':   array.name if (array.name is not None) else '',
             'attrs':  array.attrs if bool(array.attrs) else []    # Tests if dict is empty
            }

    # Extract data from variable as numpy ndarray
    array = array.values

    return array,attrs


def _parse_typemap(typemap_in=None):
    """
    Combine any input values for typemap (which maps loaded Matlab variable
    types to Python variable types) with defaults

    Parameters
    ----------
    typemap_in : dict, {str:str}
        Maps names of Matlab variable types or specific Matlab variable names
        to Python variable types.
        Matlab types: 'array' = numerical array, 'cell' = cell array, 'struct' = structure).
        Python type: 'array' = Numpy ndarray, 'dataframe' = Pandas DataFrame, 'dict' = dictionary.

    Returns
    -------
    typemap_out : dict, {str:str}, default: {'array':'array', 'cell':'array', 'struct':'dict'}
        Default mapping, overwritten with any input values.
    """
    # Set of Matlab data types to deal with here
    vbl_types = {'array', 'cell', 'struct', 'structarray'}
    # Set default values
    typemap_out = {'array':'array', 'cell':'array',
                   'struct':'dict', 'structarray':'dict'}

    # Modify defaults with any input values
    if isinstance(typemap_in,dict):
        for key in typemap_in.keys():
            # If key is in list of Matlab data types, standardize to lowercase
            # (otherwise keep as is, to match specific variable names)
            if key.lower() in vbl_types:  key = key.lower()
            typemap_out[key] = typemap_in[key].lower()

        assert typemap_out['struct'] in ['dict','dataframe'], \
            "'struct' must be mapped to 'dict' or 'dataframe' (set to %s)" \
            % typemap_out['struct']

    else:
        assert typemap_in is None, TypeError("If <typemap> is input, it must be given as a dict")

    return typemap_out


def _parse_extract_items(extract_items_in=None):
    """
    Combine any input values for extract_items (which determines whether single items in
    loaded Matlab variables are extracted from containing arrays or left as length-1 arrays)
    with defaults

    Parameters
    ----------
    extract_items_in : bool or dict, {str:bool}
        Determines whether scalar-valued Matlab variables are loaded/returned as length-1
        arrays (False) or are extracted from the containing arrays as their specific dtype (True).
        Given as a single bool value to be used for *all* loaded variables or as a
        dict that maps names of Matlab variable types or specific Matlab variable names to bools.
        Matlab types: 'array' = numerical array, 'cell' = cell array, 'struct' = structure).

    Returns
    -------
    extract_items_out : dict, {str:bool}, default: {'array':True, 'cell':False, 'struct':True}
        Default mapping, overwritten with any input values.
    """
    # Set of Matlab data types to deal with here
    vbl_types = {'array', 'cell', 'struct', 'structarray'}
    # Set default values
    extract_items_out = {'array':True, 'cell':False, 'struct':True, 'structarray':False}

    # Modify defaults with any input values
    # If input given as single bool value, set that value for all variable types
    if _isbinary(extract_items_in):
        extract_items_out = {vbl_type : extract_items_in for vbl_type in extract_items_out}

    # If input given as dict, merge with default dict
    elif isinstance(extract_items_in,dict):
        for key in extract_items_in.keys():
            assert _isbinary(extract_items_in[key]), \
                ValueError("Values for <extract_items> must be boolean (0/1)")

            # If key is in list of Matlab data types, standardize to lowercase
            # (otherwise keep as is, to match specific variable names)
            if key.lower() in vbl_types:  key = key.lower()
            extract_items_out[key] = bool(extract_items_in[key])

    else:
        assert extract_items_in is None, \
            TypeError("If <extract_items> is input, it must be given as a bool or str:bool dict")

    return extract_items_out


def _variables_to_mat(variables):
    """
    Does any necessary conversions to get all variables into matfile-compatible
    variable types format

    variables = _variables_to_mat(variables)

    Parameters
    ----------
    variables : dict, {str : <vbl>}
        Names and values of variables to convert

    Returns
    -------
    variables : dict, {str : <vbl>}
       Same as input, but with variables converted to mat compatible types:

        - xarray.DataArray -> Numpy ndarray + {string:*} dict with metadata
        attributes (stored in separate variable named <variable>_attr)

        - strings, lists of strings -> Numpy object ndarray
    """
    new_vbl_dict = {}     # In case we need to create new variables in loop below

    def _size_general(x):
        if isinstance(x,np.ndarray):        return x.size
        elif isinstance(x, (list,tuple)):   return len(x)
        else:                               return 1

    for variable,value in variables.items():
        # Call function recursively with dictionary variables
        if isinstance(value,dict):
            variables[variable] = _variables_to_mat(value)

        elif isinstance(value,list):
            # Convert lists with any strings or with unequal-length entries -> object arrays
            # (same-length numerical lists are auto-converted to numerical arrays)
            # todo should we recurse here in case of nested lists?
            is_str = np.any([isinstance(item,str) for item in value])
            sizes = [_size_general(item) for item in value]
            is_unequal_size = np.any(np.diff(sizes) != 0)
            if is_str or is_unequal_size:
                variables[variable] = np.asarray(value,dtype=object)

        # Convert strings to Numpy object arrays
        elif isinstance(value,str):
            variables[variable] = np.asarray([value],dtype=object)

        # Convert Pandas DataFrame to dict and call this func recursively on items (cols)
        elif isinstance(value,pd.DataFrame):
            variables[variable] = _variables_to_mat(value.to_dict(orient='list'))

        # Conversion from xarray -> numpy array + dict (skip if xarray not installed)
        # Extract metadata attributes as new dict variable called <variable>_attr
        elif _HAS_XARRAY and isinstance(value,xr.DataArray):
            variables[variable],new_vbl_dict[variable+'_attr'] = \
                _xarray_to_array(value)


    # Merge any newly-created variables into existing variable dict
    return {**variables, **new_vbl_dict}


# =============================================================================
# v7-specific helper functions
# =============================================================================
# NOTE: This function can't be nested in _load7 b/c _dict_to_dataframe() needs access
def _process_v7_object(obj, matlab_vbl_type=None, python_vbl_type=None, extract_item=None,
                       typemap=None, extract_items=None, transpose=False, level=0, vbl=None):
    """ Properly handles arbitrary objects loaded from v7 mat files in recursive fashion """
    level += 1

    # Matlab class objects are returned as weird objects that are not parsable. Return as None.
    if matlab_vbl_type == 'class':
        converted = None

    # Matlab scalar structs are returned as (1,1) Numpy structured arrays
    #  Convert to dict or Pandas DataFrame (or leave as structured array)
    elif matlab_vbl_type == 'struct':
        if DEBUG: print('\t'*level, "struct")
        if python_vbl_type is None: python_vbl_type = typemap['struct']
        if extract_item is None: extract_item = extract_items['struct']

        if python_vbl_type.lower() == 'dict':
            converted = _structuredarray_to_dict(obj[0,0], typemap=typemap, transpose=transpose,
                                                 extract_item=extract_item,
                                                 extract_items=extract_items, level=level)

        elif python_vbl_type.lower() == 'dataframe':
            converted = _structuredarray_to_dataframe(obj[0,0], typemap=typemap, transpose=transpose,
                                                      extract_item=extract_item,
                                                      extract_items=extract_items, level=level)

        else:
            converted = obj[0,0]
            assert python_vbl_type.lower() != 'structuredarray', \
                ValueError("%s is an unsupported output type for Matlab structs. \n \
                            Must be 'dict'|'dataFrame'|'structuredArray'" % python_vbl_type)

    # Matlab struct arrays are returned as (m,n) Numpy structured arrays
    #  Convert to dict (or leave as structured array)
    elif matlab_vbl_type == 'structarray':
        if DEBUG: print('\t'*level, "structarray")
        if python_vbl_type is None: python_vbl_type = typemap['structarray']
        if extract_item is None: extract_item = extract_items['structarray']

        if python_vbl_type.lower() == 'dict':
            converted = _structuredarray_to_dict(obj, typemap=typemap, transpose=transpose,
                                                 extract_item=extract_item,
                                                 extract_items=extract_items, level=level)

        else:
            converted = obj
            assert python_vbl_type.lower() != 'structuredarray', \
                ValueError("%s is an unsupported output type for Matlab struct arrays. \n \
                            Must be 'dict'|'structuredArray'" % python_vbl_type)

    # Cell arrays are returned as object arrays. Squeeze out any singleton axes and
    # Permute axes of each cell's contents if 'PYTHON' order requested
    elif matlab_vbl_type == 'cell':
        if DEBUG: print('\t'*level, "cell")
        if extract_item is None: extract_item = extract_items['cell']

        flatiter = obj.flat             # 1D iterator to iterate over arbitrary-shape array
        converted = np.ndarray(shape=obj.shape,dtype=object) # Create empty object array
        for _ in range(obj.size):
            coords = flatiter.coords    # Multidim coordinates into array
            # Infer Matlab variable type of cell array element
            matlab_elem_type = _v7_matlab_type(obj[coords])
            elem_c = _process_v7_object(obj[coords], typemap=typemap,
                                        matlab_vbl_type=matlab_elem_type,
                                        extract_item=extract_item, extract_items=extract_items,
                                        transpose=transpose, level=level)
            if not extract_item: elem_c = np.atleast_1d(elem_c)
            converted[coords] = elem_c
            next(flatiter)              # Iterate to next element

    # General numerical/logical and cell arrays
    elif isinstance(obj,np.ndarray):
        if DEBUG: print('\t'*level, "array")
        # Logicals (binary-valued variables): Convert to bool type
        if matlab_vbl_type == 'logical':
            if DEBUG: print('\t'*level, "logical")
            converted = obj.astype(bool)

        # Strings: Extract string (dealing with empty strings appropriately)
        elif obj.dtype.type == np.str_:
            if DEBUG: print('\t'*level, "string")
            # Empty strings: return empty str
            if len(obj) == 0:   converted = str()
            # Character arrays (not strings per se): keep as array of str's
            elif len(obj) > 1:  converted = obj
            # General strings: Convert to string type (from np.str_)
            else:               converted = str(obj[0])
            if not extract_item: converted = np.asarray(converted, dtype=object)

        # General numerical array: Copy as-is
        else:
            if DEBUG: print('\t'*level, "numerical")
            converted = obj

    # Note: Only do the following for objects that remain output as ndarrays
    if isinstance(converted,np.ndarray):
        if extract_item is None: extract_item = extract_items['array']

        # Length-1 arrays: Extract single item as type <dtype> (if <extract_item> is True)
        if ((converted.shape == ()) or (converted.size == 1)) and extract_item:
            converted = converted.item()

        else:
            # Squeeze out any singleton axes, eg: Reshape (1,n) ndarrays -> (n,) vectors
            converted = converted.squeeze()

            # Permute array axes if 'PYTHON'/'C'/'ROWMAJOR' order requested
            if transpose: converted = converted.T

    return converted


def _v7_matlab_type(obj):
    """ Infer Matlab variable type of objects loaded from v7 mat files """
    # Matlab class objects are returned as this weird object (which is unparsable)
    if isinstance(obj,scipy.io.matlab.mio5_params.MatlabOpaque):
        warn("Can't convert Matlab 'object' type (unknown proprietary data type). Returning None.")
        return 'class'
    
    # Matlab scalar structs are returned as (1,1) Numpy structured arrays
    elif _is_structured_array(obj) and (obj.size == 1): return 'struct'
    
    # Matlab struct arrays are returned as (m,n) Numpy structured arrays
    elif _is_structured_array(obj) and (obj.size > 1):  return 'structarray'
    
    elif isinstance(obj,np.ndarray):
        # Cell arrays are returned as object arrays
        if obj.dtype == object:                                         return 'cell'
        # Heuristically assume arrays with only 0/1 (but not all 0's) are logicals
        elif _isbinary(obj) and len(obj) > 1 and not (obj == 0).all():  return 'logical'
        # General numeric array
        else:                                                           return 'array'
        
    else:
        raise TypeError("Undetermined type of variable:", obj)


# =============================================================================
# v7.3-specific helper functions
# =============================================================================
def _h5py_matlab_type(obj):
    """ Return variable type of Matlab variable encoded in h5py object """
    assert 'MATLAB_class' in obj.attrs, \
        AttributeError("Can't determine Matlab variable type. " \
                       "No 'MATLAB_class' attribute in h5py object '%s'" % obj)

    # Extract attribute with Matlab variable type, convert bytes -> string
    return obj.attrs['MATLAB_class'].decode('UTF-8')


def _convert_string(value, encoding='UTF-16'):
    """ Convert integer-encoded strings in HDF5 files to strings """
    return ''.join(value[:].tobytes().decode(encoding))    # time test ~ 700 ms

    # Note: Alternative methods that tested much slower:
    # return ''.join([chr(c) for c in value])               # time test ~ 7.2 s
    # return ''.join(map(chr,value))                        # time test ~ 7.3 s



"""
Suite of tests to assess "face validity" of synchrony analysis functions in sync.py
Usually used to test new or majorly updated functions.

Includes tests that parametrically estimate synchrony as a function of frequency, amplitude, phase,
n, etc. to establish methods produce expected pattern of results.

Plots results and runs assertions that basic expected results are reproduced

Functions
---------
- test_synchrony :          Contains tests of synchrony estimation functions
- synchrony_test_battery :  Runs standard battery of tests of synchrony estimation functions
- spike_field_test_battery :Runs standard battery of tests of spike-field sync estimation functions
"""
import os
import time
from warnings import warn
from copy import deepcopy
from math import pi, ceil, floor, log2
import numpy as np
import matplotlib.pyplot as plt

from scipy.stats import bernoulli

from spynal.sync.sync import simulate_multichannel_oscillation, synchrony
from spynal.info import neural_info


# =============================================================================
# Validity test functions
# =============================================================================
def test_synchrony(method, pair_type='lfp-lfp', test='frequency', test_values=None,
                   spec_method='wavelet', single_trial=None,
                   do_tests=True, do_plots=False, plot_dir=None,
                   seed=1, phi_sd=pi/4, dphi=0, damp=1, amp=5.0, freq=32, phi=0, noise=0.5,
                   n=1000, time_range=3.0, smp_rate=1000, burst_rate=0, **kwargs):
    """
    Basic testing for functions estimating bivariate time-frequency oscillatory synchrony/coherence

    Generates synthetic LFP data using given parameters, estimates time-frequency synchrony
    using given function, and compares estimated to expected.

    For test failures, raises an error or warning (depending on value of `do_tests`).
    Optionally plots summary of test results.

    Parameters
    ----------
    method : str
        Name of synchrony estimation function to test. Options: 'PPC' | 'PLV' | 'coherence'

    test : str, default: 'frequency'
        Type of test to run. Options:
        - 'synchrony' : Tests multiple values of strength of synchrony (by manipulating phase SD
            of one signal). Checks for monotonic increase of synchrony measure.
        - 'frequency' : Tests multiple simulated oscillatory frequencies.
            Checks for monotonic increase of peak freq.
        - 'relphase' : Tests multiple simulated btwn-signal relative phases (dphi).
            Checks that synchrony doesn't vary appreciably.
        - 'ampratio' : Test multiple btwn-signal amplitude ratios (damp).
            Checks that synchrony doesn't vary appreciably.
        - 'phase' : Tests multiple simulated absolute phases (phi).
            Checks that synchrony doesn't vary appreciably.
        - 'n' : Tests multiple values of number of trials (n).
            Checks that synchrony doesn't greatly vary with n.

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:
        
        - 'synchrony' : Relative phase SD's (~inverse of synchrony). Default: [pi,pi/2,pi/4,pi/8,0]
        - 'frequency' : Frequencies to test. Default: [4,8,16,32,64]
        - 'relphase' :  Relative phases to test. Default: [-pi,-pi/2,0,pi/2,pi]
        - 'ampratio' :  Amplitude ratios to test. Default: [1,2,4,8]
        - 'amplitude' : Oscillation amplitudes to test. Default: [1,2,5,10,20]
        - 'phase' :     Absolute phases to test. Default: [-pi,-pi/2,0,pi/2,pi]
        - 'n' :         Trial numbers. Default: [25,50,100,200,400,800]

    spec_method : str, default: 'wavelet'
        Name of spectral estimation function to use to generate time-frequency representation
        to input into synchrony function

    single_trial : str or None, default: None
        What type of synchrony estimator to compute:
        - None :        standard across-trial estimator [default]
        - 'pseudo' :    single-trial estimates using jackknife pseudovalues
        - 'richter' :   single-trial estimates using actual jackknifes as in Richter & Fries 2015

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.
       
    - Following args set param's for sim, may be overridden by <test_values> depending on test -
    phi_sd  Scalar. SD (rad) for phase diff of 2 signals if test != 'synchrony'. Default: pi/2
    dphi    Scalar. Phase difference (rad) of 2 simulated signals if test != 'relphase'. Default: 0
    damp    Scalar. Amplitude ratio of 2 simulated signals. Default: 1 (same amplitude)
    amp     Scalar. Simulated oscillation amplitude (a.u.) if test != 'amplitude'. Default: 5.0
    freq    Scalar. Simulated oscillation frequency (Hz) if test != 'frequency'. Default: 32
    phi     Scalar. Simulated oscillation phase (rad). Default: 0
    noise   Scalar. Additive noise for simulated signal (a.u., same as amp). Default: 0.5
    n       Int. Number of trials to simulate if test != 'n'. Default: 1000
    time_range Scalar. Full time range to simulate oscillation over (s). Default: 3 s
    smp_rate Int. Sampling rate for simulated data (Hz). Default: 1000
    burst_rate Scalar. Oscillatory burst rate (bursts/trial). Default: 0 (non-bursty)

    **kwargs :
        All other keyword args passed to synchrony estimation function
 
    Returns
    -------
    syncs : ndarray, shape=(n_freqs,n_timepts,n_values)
        Estimated synchrony strength for each tested value

    phases : ndarray=(n_freqs,n_timepts,n_values)
        Estimated synchrony phase for each tested value

    passed : bool
        True if all tests produce expected values; otherwise False.
    """
    method = method.lower()
    test = test.lower()

    # Set defaults for tested values and set up rate generator function depending on <test>
    sim_args = dict(amplitude=[amp,amp*damp], phase=[phi+dphi,phi], phase_sd=[0,phi_sd],
                    n_trials=n, noise=noise, time_range=time_range, burst_rate=burst_rate,
                    seed=seed)

    if test in ['synchrony','strength','coupling']:
        test_values = [pi, pi/2, pi/4, 0] if test_values is None else test_values
        del sim_args['phase_sd']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda phi_sd: simulate_multichannel_oscillation(2,freq,**sim_args,
                                                                    phase_sd=[0,phi_sd])

    elif test in ['relphase','rel_phase','dphi']:
        test_values = [-pi,-pi/2,0,pi/2,pi] if test_values is None else test_values
        del sim_args['phase']   # Delete preset arg so it uses argument to lambda below
        # Note: Implement dphi in 1st channel only so synchrony phase ends up monotonically
        # *increasing* for tests below
        gen_data = lambda dphi: simulate_multichannel_oscillation(2,freq,**sim_args,
                                                                  phase=[phi+dphi,phi])

    elif test in ['ampratio','amp_ratio','damp']:
        test_values = [1,2,4,8] if test_values is None else test_values
        del sim_args['amplitude']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda damp: simulate_multichannel_oscillation(2,freq,**sim_args,
                                                                  amplitude=[amp,amp*damp])

    elif test in ['frequency','freq']:
        test_values = [4,8,16,32,64] if test_values is None else test_values
        gen_data = lambda freq: simulate_multichannel_oscillation(2,freq,**sim_args)

    elif test in ['amplitude','amp']:
        test_values = [1,2,5,10,20] if test_values is None else test_values
        del sim_args['amplitude']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda amp: simulate_multichannel_oscillation(2,freq,**sim_args,
                                                                 amplitude=[amp,amp*damp])

    elif test in ['phase','phi']:
        test_values = [-pi,-pi/2,0,pi/2,pi] if test_values is None else test_values
        del sim_args['phase']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda phi: simulate_multichannel_oscillation(2,freq,**sim_args,
                                                                 phase=[phi,phi+dphi])

    elif test in ['n','n_trials']:
        test_values = [25,50,100,200,400,800] if test_values is None else test_values
        del sim_args['n_trials']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda n: simulate_multichannel_oscillation(2,freq,**sim_args,n_trials=n)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    # Ensure hand-set values are sorted (ascending), as many tests assume it
    test_values = sorted(test_values,reverse=True) if test in ['synchrony','strength','coupling'] \
                  else sorted(test_values)
    n_values = len(test_values)

    # Set default parameters for each spectral estimation method
    if spec_method == 'multitaper':
        if 'freq_range' not in kwargs:  kwargs['freq_range'] = [1,100]
    elif spec_method == 'bandfilter':
        if 'freqs' not in kwargs:       kwargs['freqs'] = ((2,6),(6,10),(10,22),(22,42),(42,86))

    if 'buffer' not in kwargs: kwargs['buffer'] = 1.0

    for i,value in enumerate(test_values):
        # print("Running test value %d/%d: %.2f" % (i+1,n_values,value))

        # Simulate data with oscillation of given params in additive noise
        #  -> (n_timepts,n_trials,n_chnls=2)
        data = gen_data(value)
        if pair_type == 'spk-lfp':
            data[:,:,0] = _continuous_to_spiking(data[:,:,0])

        # Compute time-frequency/spectrogram representation of data and
        # bivariate measure of synchrony -> (n_freqs,n_timepts)
        if single_trial is None:
            sync,freqs,timepts,phase = synchrony(data[:,:,0], data[:,:,1], axis=1, method=method,
                                                spec_method=spec_method, smp_rate=smp_rate,
                                                time_axis=0, return_phase=True,
                                                single_trial=single_trial, **kwargs)
        else:
            sync,freqs,timepts = synchrony(data[:,:,0], data[:,:,1], axis=1, method=method,
                                           spec_method=spec_method, smp_rate=smp_rate,
                                           time_axis=0, return_phase=False,
                                           single_trial=single_trial, **kwargs)
            print(sync.min(), sync.max())
            sync = sync.mean(axis=-1)

        n_freqs,n_timepts = sync.shape
        if freqs.ndim == 2:
            bands = freqs
            freqs = freqs.mean(axis=1)  # Compute center of freq bands

        # KLUDGE Initialize output arrays on 1st loop, once spectrogram output shape is known
        if i == 0:
            syncs = np.empty((n_freqs,n_timepts,n_values))
            if single_trial is None:
                phases = np.empty((n_freqs,n_timepts,n_values))
            else:
                phases = None

        syncs[:,:,i] = sync
        if single_trial is None: phases[:,:,i] = phase

    # Compute mean (weighted circular mean for phase) across all timepoints
    # -> (n_freqs,n_values) frequency marginal for each tested value
    marginal_syncs = syncs.mean(axis=1)
    if single_trial is None:
        marginal_phases = np.angle(_amp_phase_to_complex(syncs,phases).mean(axis=1))

    # For wavelets, evaluate and plot frequency on log scale
    if spec_method == 'wavelet':
        freq_transform  = np.log2
        plot_freqs      = freq_transform(freqs)
        fmin            = ceil(log2(freqs[0]))
        fmax            = floor(log2(freqs[-1]))
        freq_ticks      = np.arange(fmin,fmax+1)
        freq_tick_labels= 2**np.arange(fmin,fmax+1)

    # For bandfilter, plot frequency bands in categorical fashion
    elif spec_method == 'bandfilter':
        freq_transform  = lambda x: np.argmin(np.abs(x - freqs))  # Index of closest sampled freq
        plot_freqs      = np.arange(len(freqs))
        freq_ticks      = np.arange(len(freqs))
        freq_tick_labels= bands

    # For other spectral analysis, evaluate and plot frequency on linear scale
    else:
        freq_transform  = lambda x: x
        plot_freqs      = freqs
        fmin            = ceil(freqs[0]/10.0)*10.0
        fmax            = floor(freqs[-1]/10.0)*10.0
        freq_ticks      = np.arange(fmin,fmax+1,10).astype(int)
        freq_tick_labels= freq_ticks

    freqs_transformed   = np.asarray([freq_transform(f) for f in freqs])

    # For frequency test, find frequency with maximal synchrony for each test
    if test in ['frequency','freq']:
        idxs = np.argmax(marginal_syncs,axis=0)
        peak_freqs = freqs[idxs] if spec_method != 'bandfilter' else idxs

        # Find frequency in spectrogram closest to each simulated frequency
        test_freq_idxs  = np.asarray([np.argmin(np.abs(freq_transform(f) - freqs_transformed))
                                      for f in test_values])

        # Extract synchrony and phase at each tested frequency
        test_freq_syncs = marginal_syncs[test_freq_idxs,np.arange(n_values)]
        if single_trial is None:
            test_freq_phases = marginal_phases[test_freq_idxs,np.arange(n_values)]

    else:
        # Find frequency in spectrogram closest to simulated frequency
        test_freq_idx  = np.argmin(np.abs(freq_transform(freq) - freqs_transformed))

        # Extract synchrony and phase at simulated frequency
        test_freq_syncs = marginal_syncs[test_freq_idx,:]
        if single_trial is None:
            test_freq_phases = marginal_phases[test_freq_idx,:]

    if do_plots:
        dt      = np.diff(timepts).mean()
        tlim    = [timepts[0]-dt/2, timepts[-1]+dt/2]
        df      = np.diff(plot_freqs).mean()
        flim    = [plot_freqs[0]-df/2, plot_freqs[-1]+df/2]
        plot_variables = ['sync','phase'] if single_trial is None else ['sync']

        # # Plot synchrony/phase spectrogram for each tested value
        # n_subplots = [floor(n_values/2), ceil(n_values/floor(n_values/2))]
        # for i_vbl,variable in enumerate(plot_variables):
        #     plot_vals = syncs if variable == 'sync' else phases
        #     cmap = 'viridis' if variable == 'sync' else 'hsv'
        #     plt.figure()
        #     for i,value in enumerate(test_values):
        #         clim = [plot_vals[:,:,i].min(),plot_vals[:,:,i].max()] if variable == 'sync' \
        #                else [-pi,pi]
        #         ax = plt.subplot(n_subplots[0],n_subplots[1],i+1)
        #         plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        #         target_freq = freq_transform(value) if test in ['frequency','freq'] else \
        #                       freq_transform(freq)
        #         plt.plot(tlim, [target_freq,target_freq], '-', color='r', linewidth=0.5)
        #         plt.imshow(plot_vals[:,:,i], extent=[*tlim,*flim], vmin=clim[0], vmax=clim[1],
        #                    aspect='auto', origin='lower', cmap=cmap)
        #         if i in [0,n_subplots[1]]:
        #             plt.yticks(freq_ticks,freq_tick_labels)
        #         else:
        #             ax.set_xticklabels([])
        #             plt.yticks(freq_ticks,[])
        #         plt.title(np.round(value,decimals=2))
        #         plt.colorbar()
        #     plt.show()
        # if plot_dir is not None:
        #   plt.savefig(os.path.join(plot_dir,'synchrony-spectrogram-%s-%s-%s.png'
        #               % (method,test,spec_method)))


        # Plot time-averaged synchrony/phase spectrum for each tested value
        plt.figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        threshold_syncs = marginal_syncs > 0.1

        for i_vbl,variable in enumerate(plot_variables):
            plt.subplot(1,2,i_vbl+1)
            plot_vals = marginal_syncs if variable == 'sync' else marginal_phases
            ylim = [0,1.05*marginal_syncs.max()] if variable == 'sync' else [-pi,pi]
            for i,value in enumerate(test_values):
                if variable == 'phase':
                    plt.plot(plot_freqs, plot_vals[:,i], '-', color=colors[i],
                             linewidth=1.5, alpha=0.33)
                    plt.plot(plot_freqs[threshold_syncs[:,i]], plot_vals[threshold_syncs[:,i],i],
                             '.-', color=colors[i], linewidth=1.5)
                else:
                    plt.plot(plot_freqs, plot_vals[:,i], '.-', color=colors[i], linewidth=1.5)
                target_freq = freq_transform(value) if test in ['frequency','freq'] else \
                              freq_transform(freq)
                plt.plot([target_freq,target_freq], ylim, '-', color=colors[i], linewidth=0.5)
                plt.text(flim[1]-0.05*np.diff(flim), ylim[1]-(i+1)*0.05*np.diff(ylim),
                         np.round(value,decimals=2), color=colors[i], fontweight='bold',
                         horizontalalignment='right')
            plt.xlim(flim)
            plt.ylim(ylim)
            plt.xticks(freq_ticks,freq_tick_labels)
            plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
            plt.xlabel('Frequency (Hz)')
            plt.ylabel(variable)
            if i_vbl == 0:
                title = "%s %s %s test" % (spec_method,method,test)
                if single_trial is not None: title += " (%s trials)" % single_trial
                plt.title(title, horizontalalignment='left')
        plt.show()
        if plot_dir is not None:
            filename = 'synchrony-spectrum-%s-%s-%s' % (method,test,spec_method)
            if single_trial is not None: filename += '-'+single_trial
            plt.savefig(os.path.join(plot_dir, filename+'.png'))

        # Plot summary curve of synchrony (or peak frequency) vs tested value
        plt.figure()
        for i_vbl,variable in enumerate(plot_variables):
            ax = plt.subplot(1,2,i_vbl+1)
            plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
            if test in ['frequency','freq'] and variable == 'sync':
                lim = (0,1.1*freq_transform(test_values[-1]))
                plt.plot(lim, lim, color='k', linewidth=0.5)
                if spec_method == 'bandfilter':
                    plt.plot([freq_transform(f) for f in test_values], peak_freqs, marker='o')
                else:
                    plt.plot([freq_transform(f) for f in test_values],
                             [freq_transform(f) for f in peak_freqs], marker='o')
                plt.xticks(freq_ticks,freq_tick_labels)
                plt.yticks(freq_ticks,freq_tick_labels)
                plt.xlim(lim)
                plt.ylim(lim)
                ax.set_aspect('equal', 'box')
            else:
                test_freq_results = test_freq_syncs if variable == 'sync' else test_freq_phases
                xplot = [freq_transform(f) for f in test_values] \
                        if spec_method == 'bandfilter' and test in ['frequency','freq'] \
                        else test_values
                ylim = [0,1.05*test_freq_syncs.max()] if variable == 'sync' else [-pi,pi]
                plt.plot(xplot, test_freq_results, marker='o')
                plt.ylim(ylim)
            plt.xlabel('Phase SD' if test == 'synchrony' else test)
            plt.ylabel('frequency' if test in ['frequency','freq'] and variable == 'sync' else \
                        variable)
            if i_vbl == 0:
                plt.title("%s %s %s test" % (spec_method,method,test), horizontalalignment='left')
        plt.show()
        if plot_dir is not None:
            filename = 'synchrony-summary-%s-%s-%s.png' % (method,test,spec_method)
            if single_trial is not None: filename += '-'+single_trial
            plt.savefig(os.path.join(plot_dir, filename+'.png'))

    ## Determine if test actually produced the expected values
    # 'synchrony' : Test if synchrony strength increases monotonically with simulated synchrony
    if test in ['synchrony','strength','coupling']:
        evals = [((np.diff(test_freq_syncs) > 0).all(),
                    "Estimated sync strength does not increase monotonically with simulated sync")]

    # 'frequency' : check if frequency of peak synchrony matches simulated target frequency
    elif test in ['frequency','freq']:
        evals = [((np.diff(peak_freqs) > 0).all(),
                    "Estimated peak freq does not increase monotonically with expected freq")]

    # 'amplitude','phase','ampratio' : Test if synchrony is ~ same for all values
    elif test in ['amplitude','amp', 'phase','phi', 'ampratio','amp_ratio','damp']:
        evals = [(test_freq_syncs.ptp() < 0.1,
                    "Estimated sync has larger than expected range across tested %s value" % test)]

    # 'relphase' : Test if sync strength is ~ same for all values, phase increases monotonically
    elif test in ['relphase','rel_phase','dphi']:
        circ_subtract = lambda data1,data2: np.angle(np.exp(1j*data1) / np.exp(1j*data2))
        circ_diff = lambda data: circ_subtract(data[1:],data[:-1])

        evals = [(test_freq_syncs.ptp() < 0.1,
                    "Estimated sync has larger than expected range across tested %s value" % test)]
        if single_trial is None:
            evals.append(((circ_diff(test_freq_phases) > 0).all(),
                         "Estimated sync phase does not increase monotonic with sim'd relative phase"))

    # 'n' : Test if synchrony is ~ same for all values of n (unbiased by n)
    elif test in ['n','n_trials']:
        evals = [(test_freq_syncs.ptp() < 0.1,
                    "Estimated sync has > expected range across n's (likely biased by n)")]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return syncs, phases, passed


def test_sync_info(sync_method, pair_type='lfp-lfp', test='sync', test_values=None,
                   spec_method='wavelet', single_trial_method='pseudo', info_method='pev',
                   do_tests=True, do_plots=False, plot_dir=None, seed=1,
                   phi_sd=pi/2, dphi=0, damp=1, amp=5.0, freq=32, phi=0, noise=0.5,
                   n=50, time_range=3.0, smp_rate=1000, burst_rate=0, **kwargs):
    """
    Basic testing for information contained in single-trial synchrony estimates

    Generates synthetic LFP data using given parameters, with some difference across
    stimulated conditions, computes single-trial (jackknife) time-frequency synchrony estimates
    using given function, computes measures of information on those, and compares to expected.

    Parameters
    ----------
    sync_method : str
        Name of synchrony estimation function to test. Options: 'PPC' | 'PLV' | 'coherence'

    test : str, default: 'frequency'
        Type of test to run. Options:
        
        - 'synchrony' : Tests multiple values of btwn-condition difference in synchrony strength
            (by manipulating phase SD of one signal). Checks for monotonic increase of information.
        - 'relphase' : Tests multiple values of btwn-condition difference in relative phase.
            Checks that information doesn't vary appreciably.
        - 'ampratio' : Test multiple values of btwn-condition difference in relative amplitude.
            Checks that information doesn't vary appreciably.

    test_values : array-like, shape=(n_values,), dtype=str
        List of values to test. Interpretation and defaults are test-specific:
        
        - 'synchrony' : Relative phase SDs (~inverse of synchrony strength). Default: [pi,pi/2,pi/4,pi/8,0]
        - 'relphase' : Relative phases to test. Default: [-pi,-pi/2,0,pi/2,pi]
        - 'ampratio' : Amplitude ratios to test. Default: [1,2,4,8]

    spec_method : str, default: 'wavelet'
        Name of spectral estimation function to use to generate time-frequency representation
        to input into synchrony function. Options: 'wavelet' | 'multitaper' | 'bandfilter'

    single_trial_method : str, default: 'pseudo'
        What type of single-trial synchrony estimator to compute:
        
        - 'pseudo' : single-trial estimates using jackknife pseudovalues
        - 'richter' : single-trial estimates using actual jackknifes as in Richter & Fries 2015

    do_tests : bool, default: True
        Set=True to evaluate test results against expected values and raise an error if they fail

    do_plots : bool, default: False
        Set=True to plot test results

    plot_dir : str, default: None (don't save to file)
        Full-path directory to save plots to. Set=None to not save plots.

    seed : int, default: 1 (reproducible random numbers)
        Random generator seed for repeatable results. Set=None for fully random numbers.

    - Following args set param's for sim, may be overridden by <test_values> depending on test -
    phi_sd  Scalar. SD (rad) for phase diff of 2 signals if test != 'synchrony'. Default: pi/2
    dphi    Scalar. Phase difference (rad) of 2 simulated signals if test != 'relphase'. Default: 0
    damp    Scalar. Amplitude ratio of 2 simulated signals. Default: 1 (same amplitude)
    amp     Scalar. Simulated oscillation amplitude (a.u.) if test != 'amplitude'. Default: 5.0
    freq    Scalar. Simulated oscillation frequency (Hz) if test != 'frequency'. Default: 32
    phi     Scalar. Simulated oscillation phase (rad). Default: 0
    noise   Scalar. Additive noise for simulated signal (a.u., same as amp). Default: 0.5
    n       Int. Number of trials to simulate if test != 'n'. Default: 1000
    time_range Scalar. Full time range to simulate oscillation over (s). Default: 3 s
    smp_rate Int. Sampling rate for simulated data (Hz). Default: 1000
    burst_rate Scalar. Oscillatory burst rate (bursts/trial). Default: 0 (non-bursty)

    **kwargs :
        All other keyword args passed to synchrony estimation function

    Returns
    -------
    infos : ndarray, shape=(n_freqs,n_timepts,n_values)
        Estimated neural information in synchrony.

    syncs : ndarray, shape=(n_freqs,n_timepts,n_trials,n_values)
        Estimated single-trial synchrony for each tested value.

    passed : bool
        True if all tests produce expected values; otherwise False.
    """
    test = test.lower()
    sync_method = sync_method.lower()

    # Set defaults for tested values and set up rate generator function depending on <test>
    sim_args = dict(amplitude=[amp,amp*damp], phase=[phi+dphi,phi], phase_sd=[0,phi_sd],
                    n_trials=n, noise=noise, time_range=time_range, burst_rate=burst_rate,
                    seed=seed)

    def simulate_synchrony_information(test_param, base_value, test_value, freq, **sim_args):
        """ Generate synthetic multichannel oscillatory data with diff params across conds """
        base_args = deepcopy(sim_args)
        test_args = deepcopy(sim_args)

        # Set value for manipulated parameter in 'baseline' and 'test' condition trials
        base_args[test_param] = base_value
        test_args[test_param] = test_value

        # Generate simulated oscillations for baseline and test condition trials,
        # concatenate alog trial axis
        return np.concatenate((simulate_multichannel_oscillation(2,freq,**base_args),
                               simulate_multichannel_oscillation(2,freq,**test_args)),
                              axis=1)

    if test in ['sync','synchrony','strength','coupling']:
        base_value = sim_args['phase_sd']
        test_values = [pi/2, pi/4, pi/8, 0] if test_values is None else test_values
        del sim_args['phase_sd']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda phi_sd: simulate_synchrony_information('phase_sd', base_value, [0,phi_sd],
                                                                 freq, **sim_args)

    elif test in ['relphase','rel_phase','dphi']:
        base_value = sim_args['phase']
        test_values = [-pi,-pi/2,0,pi/2,pi] if test_values is None else test_values
        del sim_args['phase']   # Delete preset arg so it uses argument to lambda below
        # Note: Implement dphi in 1st channel only so synchrony phase ends up monotonically
        # *increasing* for tests below
        gen_data = lambda dphi: simulate_synchrony_information('phase', base_value, [phi+dphi,phi],
                                                               freq, **sim_args)

    elif test in ['ampratio','amp_ratio','damp']:
        base_value = sim_args['amplitude']
        test_values = [1,2,4,8] if test_values is None else test_values
        del sim_args['amplitude']   # Delete preset arg so it uses argument to lambda below
        gen_data = lambda damp: simulate_synchrony_information('amplitude', base_value, [amp,amp*damp],
                                                               freq, **sim_args)

    else:
        raise ValueError("Unsupported value '%s' set for <test>" % test)

    labels = np.hstack((np.zeros((n,)),np.ones((n,))))

    # Ensure hand-set values are sorted (ascending), as many tests assume it
    test_values = sorted(test_values,reverse=True) if test in ['synchrony','strength','coupling'] \
                  else sorted(test_values)
    n_values = len(test_values)

    # Set default parameters for each spectral estimation method
    if spec_method == 'multitaper':
        if 'freq_range' not in kwargs:  kwargs['freq_range'] = [1,100]
    elif spec_method == 'bandfilter':
        if 'freqs' not in kwargs:       kwargs['freqs'] = ((2,6),(6,10),(10,22),(22,42),(42,86))

    if ('buffer' not in kwargs) and (spec_method != 'multitaper'): kwargs['buffer'] = 1.0
    # TEMP if 'buffer' not in kwargs: kwargs['buffer'] = 1.0

    for i,value in enumerate(test_values):
        # print("Running test value %d/%d: %.2f" % (i+1,n_values,value))

        # Simulate data with oscillation of given params in additive noise
        #  -> (n_timepts,n_trials,n_chnls=2)
        data = gen_data(value)
        if pair_type == 'spk-lfp':
            data[:,:,0] = _continuous_to_spiking(data[:,:,0])

        # Compute single-trial time-frequency/spectrogram representation of data and
        # bivariate measure of synchrony -> (n_freqs,n_timepts,n_trials)
        sync,freqs,timepts = synchrony(data[:,:,0], data[:,:,1], axis=1, method=sync_method,
                                        spec_method=spec_method, smp_rate=smp_rate,
                                        time_axis=0, return_phase=False,
                                        single_trial=single_trial_method, keepdims=False,
                                        **kwargs)

        n_freqs,n_timepts,_ = sync.shape
        if freqs.ndim == 2:
            bands = freqs
            freqs = freqs.mean(axis=1)  # Compute center of freq bands

        # Compute neural information about condition in synchrony
        info = neural_info(labels, sync, axis=-1, method=info_method)

        # KLUDGE Initialize output arrays on 1st loop, once spectrogram output shape is known
        if i == 0:
            syncs = np.empty((n_freqs,n_timepts,n*2,n_values))
            infos = np.empty((n_freqs,n_timepts,n_values))

        syncs[:,:,:,i] = sync
        infos[:,:,i] = info.squeeze(axis=-1)

    # Take marginal means across time -> (freq, [trials], values)
    marginal_syncs = syncs.mean(axis=1)
    marginal_infos = infos.mean(axis=1)

    # For wavelets, evaluate and plot frequency on log scale
    if spec_method == 'wavelet':
        freq_transform  = np.log2
        plot_freqs      = freq_transform(freqs)
        fmin            = ceil(log2(freqs[0]))
        fmax            = floor(log2(freqs[-1]))
        freq_ticks      = np.arange(fmin,fmax+1)
        freq_tick_labels= 2**np.arange(fmin,fmax+1)

    # For bandfilter, plot frequency bands in categorical fashion
    elif spec_method == 'bandfilter':
        freq_transform  = lambda x: np.argmin(np.abs(x - freqs))  # Index of closest sampled freq
        plot_freqs      = np.arange(len(freqs))
        freq_ticks      = np.arange(len(freqs))
        freq_tick_labels= bands

    # For other spectral analysis, evaluate and plot frequency on linear scale
    else:
        freq_transform  = lambda x: x
        plot_freqs      = freqs
        fmin            = ceil(freqs[0]/10.0)*10.0
        fmax            = floor(freqs[-1]/10.0)*10.0
        freq_ticks      = np.arange(fmin,fmax+1,10).astype(int)
        freq_tick_labels= freq_ticks

    freqs_transformed   = np.asarray([freq_transform(f) for f in freqs])

    # Find frequency in spectrogram closest to simulated frequency
    test_freq_idx   = np.argmin(np.abs(freq_transform(freq) - freqs_transformed))
    # Extract synchrony and phase at simulated frequency
    test_freq_infos = marginal_infos[test_freq_idx,:]


    if do_plots:
        dt      = np.diff(timepts).mean()
        tlim    = [timepts[0]-dt/2, timepts[-1]+dt/2]
        df      = np.diff(plot_freqs).mean()
        flim    = [plot_freqs[0]-df/2, plot_freqs[-1]+df/2]

        # Plot time-averaged synchrony spectrum and neural information for each tested value
        plt.figure()
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

        # Plot synchrony spectrum (at tested freq) as a function of trials for each test value
        plt.subplot(1,2,1)
        for i,value in enumerate(test_values):
            plt.plot(np.arange(n*2)+1, marginal_syncs[test_freq_idx,:,i], '.-',
                     color=colors[i], linewidth=1.5)
        plt.xlim((1,n*2))
        plt.xlabel('Trials')
        plt.ylabel(single_trial_method + ' ' + sync_method)

        # Plot neural information spectrum for each test value
        plt.subplot(1,2,2)
        ylim = (marginal_infos.min(), marginal_infos.max())
        ylim = (min(0, ylim[0]-0.05*(ylim[1]-ylim[0])), ylim[1]+0.05*(ylim[1]-ylim[0]))
        for i,value in enumerate(test_values):
            plt.plot(plot_freqs, marginal_infos[:,i], '.-', color=colors[i], linewidth=1.5)
            target_freq = freq_transform(value) if test in ['frequency','freq'] else \
                          freq_transform(freq)
            plt.plot([target_freq,target_freq], ylim, '-', color=colors[i], linewidth=0.5)
            plt.text(flim[1]-0.05*np.diff(flim), ylim[1]-(i+1)*0.05*np.diff(ylim),
                        np.round(value,decimals=2), color=colors[i], fontweight='bold',
                        horizontalalignment='right')
        plt.xlim(flim)
        plt.ylim(ylim)
        plt.xticks(freq_ticks,freq_tick_labels)
        plt.grid(axis='both',color=[0.75,0.75,0.75],linestyle=':')
        plt.xlabel('Frequency (Hz)')
        plt.ylabel(info_method)
        title = "%s %s %s %s test" % (spec_method,sync_method,single_trial_method,test)
        plt.title(title, horizontalalignment='left')
        plt.show()
        if plot_dir is not None:
            filename = 'info-spectrum-%s-%s-%s-%s' % (sync_method,test,spec_method,single_trial_method)
            plt.savefig(os.path.join(plot_dir, filename+'.png'))


    ## Determine if test actually produced the expected values
    # 'synchrony' : Test if information increases monotonically with simulated cond synchrony diff
    if test in ['sync','synchrony','strength','coupling']:
        evals = [((np.diff(test_freq_infos) > 0).all(),
                    "Estimated information does not increase monotonically with simulated sync diff")]

    # 'relphase','ampratio' : Test if information is ~ same for all values
    elif test in ['relphase','rel_phase','dphi', 'ampratio','amp_ratio','damp']:
        evals = [(test_freq_infos.ptp() < 0.1,
                    "Estimated information has larger than expected range across tested %s value" % test)]

    passed = True
    for cond,message in evals:
        if not cond:    passed = False

        # Raise an error for test fails if do_tests is True
        if do_tests:    assert cond, AssertionError(message)
        # Just issue a warning for test fails if do_tests is False
        elif not cond:  warn(message)

    return infos, syncs, passed


# =============================================================================
# Functions that loop thru an entire battery of tests
# =============================================================================
def synchrony_test_battery(methods=('PPC','PLV','coherence'),
                           tests=('synchrony','relphase','ampratio','frequency','amplitude','phase','n'),
                           spec_methods=('wavelet','multitaper','bandfilter'),
                           trial_methods=(None,'richter','pseudo'),
                           do_tests=True, **kwargs):
    """
    Run a battery of given tests on given oscillatory synchrony computation methods

    Parameters
    ----------
    methods : array-like of str, default: ('PPC','PLV','coherence') (all supported methods)
        List of synchrony computation methods to test.
                
    tests : array-like of str, default: ('synchrony','relphase','ampratio','frequency','amplitude','phase','n')
        List of tests to run. Note: certain combinations of methods,tests are skipped, 
        as they are not expected to pass (ie 'ampratio' tests skipped for coherence method)
                
    spec_methods : array-like of str, default: ('wavelet','multitaper','bandfilter')
        List of underlying spectral analysis methods to test.                

    trial_methods : array-like of str or None, default: (None,'richter','pseudo')
        List of single-trial estimate methods to test (set=None to run
        usual trial-reduced synchrony methods instead of single-trial).                

    do_tests : bool. Set=True to evaluate test results against expected values and
                raise an error if they fail. Default: True

    **kwargs :
        All other keyword args passed to synchrony estimation function
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]
    if isinstance(spec_methods,str): spec_methods = [spec_methods]
    if isinstance(trial_methods,str): trial_methods = [trial_methods]
    tests = [test.lower() for test in tests]
    methods = [method.lower() for method in methods]

    for test in tests:
        for method in methods:
            for spec_method in spec_methods:
                for trial_method in trial_methods:
                    t1 = time.time()
                    # Skip tests expected to fail due to properties of given info measures
                    # (eg ones that are biased/affected by n)
                    if (test in ['n','n_trials']) and (method in ['coherence','coh','plv']):
                        do_tests_ = False
                    elif (test in ['ampratio','amp_ratio','damp']) and (method in ['coherence','coh']):
                        do_tests_ = False
                    else:
                        do_tests_ = do_tests

                    print("Running %s test on %s %s" % (test,spec_method,method))

                    _,_,passed = test_synchrony(method, pair_type='lfp-lfp', test=test,
                                                spec_method=spec_method, single_trial=trial_method,
                                                do_tests=do_tests_, **kwargs)

                    print('%s (test ran in %.1f s)'
                        % ('PASSED' if passed else 'FAILED', time.time()-t1))

                    # If saving plots to file, let's not leave them all open
                    if 'plot_dir' in kwargs: plt.close('all')


def spike_field_test_battery(methods=('PPC','PLV','coherence'),
                           tests=('synchrony','relphase','ampratio','frequency','amplitude','phase','n'),
                           spec_methods=('wavelet','multitaper','bandfilter'),
                           do_tests=True, **kwargs):
    """
    Run a battery of given tests on given oscillatory spike-field coupling computation methods

    Parameters
    ----------
    methods : array-like of str, default: ('PPC','PLV','coherence') (all supported methods)
        List of synchrony computation methods to test.
                
    tests : array-like of str, default: ('synchrony','relphase','frequency','amplitude','phase','n')
        List of tests to run. Note: certain combinations of methods,tests are skipped, 
        as they are not expected to pass (ie 'ampratio' tests skipped for coherence method)
                
    spec_methods : array-like of str, default: ('wavelet','multitaper','bandfilter')
        List of underlying spectral analysis methods to test.                

    do_tests : bool. Set=True to evaluate test results against expected values and
                raise an error if they fail. Default: True

    **kwargs :
        All other keyword args passed to test_synchrony() function
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]
    if isinstance(spec_methods,str): spec_methods = [spec_methods]
    tests = [test.lower() for test in tests]
    methods = [method.lower() for method in methods]

    for test in tests:
        for method in methods:
            for spec_method in spec_methods:
                print("Running %s test on %s %s" % (test,spec_method,method))
                t1 = time.time()
                # Skip tests expected to fail due to properties of given info measures
                # (eg ones that are biased/affected by n)
                if (test in ['n','n_trials']) and (method in ['coherence','coh','plv']):
                    do_tests_ = False
                elif (test in ['ampratio','amp_ratio','damp']) and (method in ['coherence','coh']):
                    do_tests_ = False
                else:
                    do_tests_ = do_tests

                _,_,passed = test_synchrony(method, pair_type='spk-lfp', test=test,
                                            spec_method=spec_method, do_tests=do_tests_, **kwargs)

                print('%s (test ran in %.1f s)'
                      % ('PASSED' if passed else 'FAILED', time.time()-t1))

                # If saving plots to file, let's not leave them all open
                if 'plot_dir' in kwargs: plt.close('all')


def sync_info_test_battery(methods=('PPC','PLV','coherence'),
                           tests=('synchrony','relphase','ampratio'),
                           spec_methods=('wavelet','multitaper','bandfilter'),
                           trial_methods=('richter','pseudo'),
                           do_tests=True, **kwargs):
    """
    Run a battery of given tests on given oscillatory synchrony computation methods

    Parameters
    ----------
    methods : array-like of str, default: ('PPC','PLV','coherence') (all supported methods)
        List of synchrony computation methods to test.
                
    tests : array-like of str, default: ('synchrony','relphase','ampratio')
        List of tests to run. Note: certain combinations of methods,tests are skipped, 
        as they are not expected to pass (ie 'ampratio' tests skipped for coherence method)
                
    spec_methods : array-like of str, default: ('wavelet','multitaper','bandfilter')
        List of underlying spectral analysis methods to test.                

    trial_methods : array-like of str or None, default: (None,'richter','pseudo')
        List of single-trial estimate methods to test (set=None to run
        usual trial-reduced synchrony methods instead of single-trial).                

    do_tests : bool. Set=True to evaluate test results against expected values and
                raise an error if they fail. Default: True

    **kwargs :
        All other keyword args passed to synchrony estimation function
            
    **kwargs :
        Any other keyword passed directly to test_sync_info()
    """
    if isinstance(methods,str): methods = [methods]
    if isinstance(tests,str): tests = [tests]
    if isinstance(spec_methods,str): spec_methods = [spec_methods]
    if isinstance(trial_methods,str): trial_methods = [trial_methods]
    tests = [test.lower() for test in tests]
    methods = [method.lower() for method in methods]

    for test in tests:
        for method in methods:
            for spec_method in spec_methods:
                for trial_method in trial_methods:
                    t1 = time.time()
                    print("Running %s test on %s %s" % (test,spec_method,method))

                    _,_,passed = test_sync_info(method, pair_type='lfp-lfp', test=test,
                                                spec_method=spec_method,
                                                single_trial_method=trial_method,
                                                do_tests=do_tests, **kwargs)

                    print('%s (test ran in %.1f s)'
                        % ('PASSED' if passed else 'FAILED', time.time()-t1))

                    # If saving plots to file, let's not leave them all open
                    if 'plot_dir' in kwargs: plt.close('all')


# =============================================================================
# Helper functions
# =============================================================================
def _amp_phase_to_complex(amp,theta):
    """ Convert amplitude and phase angle to complex variable """
    return amp * np.exp(1j*theta)


def _continuous_to_spiking(data):
    """ Convert continuous data to thresholded spiking response """
    # Convert continuous oscillation to probability (range 0-1)
    data = (data - data.min()) / data.ptp()
    data = data**2  # Sparsen high rates some

    # Use probabilities to generate Bernoulli random variable at each time point
    return bernoulli.ppf(0.5, data).astype(bool)



# -*- coding: utf-8 -*-
"""
Analysis of oscillatory neural synchrony

This module is the base module for the `sync` module. It contains all the
high-level API functions. Lower-level functions can be found in the other sync modules.

Overview
--------
Contains functionality for computing oscillatory neural synchrony between pairs of 
continuous signals (LFPs/EEGs/etc.) or between pairs of spiking and continuous signals
(ie spike-LFP pair).

Currently two types of synchrony methods, based on different theoretical frameworks, are available.
Coherence is based on the theory of linear stochastic processes, and is a frequency-domain analog
of standard (Pearson) correlation. Phase-based methods (phase-locking value, pairwise phase consistency)
are based on circular statistics, and are closely related to the length of the vector average of the
relative phases of a signal pair (eg phase difference between LFP pair or spike-triggered LFP phase).

Input data (for most functions) can either be raw data or data already spectrally-transformed
(using either functionality in the `spynal.spectra` module or users' own custom code). In either
case, any synchrony method can be combined with any underlying spectral method (eg wavelet, 
multitaper, or bandfilter).
 
NOTE: Unlike other spynal functions that can run in parallel across multiple channels
("mass-univariate" analysis), these functions are currently set up to compute synchrony
on only a single pair of channels (2 LFPs or spike/LFP pair) at once.

Nevertheless, sync functions can perform operations in a mass-bivariate manner over all other
dimension. This means that rather than embedding function calls in for loops over frequencies,
timepoints, etc., like this::

    for f in frequencies:
        for t in timepoints:
            results[f,t] = compute_something(data[f,t],data2[f,t])

You can instead execute a single call on ALL the data, labeling the relevant axis
for the computation (usually trials/observations here), and it will run in parallel (vectorized)
across all frequencies, timepoints, etc. in the data, like this:

``results = compute_something(data1, data2, axis)``

Function list
-------------
General synchrony functions
^^^^^^^^^^^^^^^^^^^^^^^^^^^
- synchrony :               General synchrony between pair of analog channels using given method
- spike_field_coupling :    General spike-field coupling/synchrony btwn spike/LFP pair
- simulate_multichannel_oscillation : Generates simulated oscillatory paired data

Coherence analysis (**sync.coherence**)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- coherence :               Time-frequency coherence between pair of analog channels
- spike_field_coherence :   Spike-field coherence between a spike/LFP pair
- ztransform_coherence  :   Z-transform coherence so ~ normally distributed

Phase-based synchrony analysis (**sync.phasesync**)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
- plv :                     Phase locking value (PLV) between pair of analog channels
- ppc :                     Pairwise phase consistency (PPC) btwn pair of analog channels
- spike_field_plv :         Spike-field PLV between a spike/LFP pair
- spike_field_ppc :         Spike-field PPC between a spike/LFP pair


Function reference
------------------
"""
# Created on Thu Oct  4 15:28:15 2018
#
# @author: sbrincat

import numpy as np

from spynal.utils import set_random_seed
from spynal.spectra.utils import simulate_oscillation
from spynal.randstats.sampling import jackknifes
from spynal.sync.coherence import coherence, spike_field_coherence
from spynal.sync.phasesync import plv, ppc, spike_field_plv, spike_field_ppc


# =============================================================================
# Field-Field Synchrony functions
# =============================================================================
def synchrony(data1, data2, axis=0, method='PPC', return_phase=False, single_trial=None,
              data_type=None, spec_method='wavelet', smp_rate=None, time_axis=None,
              taper_axis=None, keepdims=True, **kwargs):
    """
    Compute measure of synchrony between pair of channels of continuous (eg LFP)
    raw or spectral (time-frequency) data, using given estimation method

    Parameters
    ----------
    data1,data2 : ndarray, shape=(...,n_obs,...), dtype=float or complex
        Single-channel continuous (eg LFP) data for 2 distinct channels.
        Can be given as raw LFPs or complex-valued time-frequency transform.

        Trial/observation axis is assumed to be axis 0 unless given in `axis`.
        For raw data, axis corresponding to time must be given in `time_axis`.

        Other than those constraints, data can have arbitrary shape, with
        analysis performed in mass-bivariate fashion independently
        along each dimension other than observation `axis` (eg different
        frequencies, timepoints, conditions, etc.)

    axis : scalar, default: 0
        Axis corresponding to distinct observations/trials

    method : {'PPC','PLV','coherence'}, default: 'PPC'
        Synchrony estimation method:

        - 'PPC' : Pairwise Phase Consistency, measure of phase synchrony (see :func:`.ppc`)
        - 'PLV' : Phase Locking Value, measure of phase synchrony (see :func:`.plv`)
        - 'coherence' : coherence, measure of linear oscillatory coupling (see :func:`.coherence`)

    return_phase : bool, default: False
        If False, only return measure of synchrony magnitude/strength between data1 and data2.
        If True, also returns mean phase difference (or coherence phase) between
        data1 and data2 (in radians) in additional output.

    single_trial : str or None, default: None
        What type of coherence estimator to compute:

        - None :       standard across-trial estimator
        - 'pseudo' :   single-trial estimates using jackknife pseudovalues
        - 'richter' :  single-trial estimates using actual jackknife estimates cf. Richter 2015

    data_type : {'raw','spectral'}, default: assume 'raw' if data is real; 'spectral' if complex
        What kind of data is input: 'raw' or 'spectral'?

    spec_method : {'wavelet','multitaper','bandfilter'}, default: 'wavelet'
        Method to use for spectral analysis (only used for *raw* data)

    smp_rate : scalar
        Sampling rate of data (only needed for *raw* data)

    time_axis : int
        Axis of data corresponding to time (only needed for *raw* data)

    taper_axis : int
        Axis of spectral data corresponding to tapers (only needed for *multitaper spectral* data)

    keepdims : bool, default: True
        If True, retains reduced trial and/or taper axes as length-one axes in output.
        If False, removes reduced trial,taper axes from outputs.

    **kwargs :
        All other kwargs passed as-is to specific synchrony estimation function.

    Returns
    -------
    sync : ndarray, shape=
        Synchrony magnitude/strength (PPC/PLV/coherence) between data1 and data2.
        If data is spectral, this has same shape as data, but with `axis` reduced
        to a singleton or removed, depending on value of `keepdims`.
        If data is raw, this has same shape as data with `axis` reduced or removed
        and a new frequency axis inserted immediately before `time_axis`.

    freqs : ndarray, shape=(n_freqs,)
        List of frequencies in `sync`. Only returned for raw data, [] otherwise.

    timepts : ndarray, shape=(n_timepts_out,)
        List of timepoints in `sync` (in s, referenced to start of data).
        Only returned for raw data, [] otherwise.

    dphi : ndarray, optional
        Mean phase difference (or coherence phase) between data1 and data2 in radians.
        Same shape as `sync`.
        Positive values correspond to data1 leading data2.
        Negative values correspond to data1 lagging behind data2.
        Optional: Only returned if return_phase is True.

    Examples
    --------
    sync, freqs, timepts = synchrony(data1, data2, axis=0, method='PPC', return_phase=False)

    sync, freqs, timepts, dphi = synchrony(data1, data2, axis=0, method='PPC', return_phase=True)

    References
    ----------
    - Single-trial method:    Womelsdorf et al (2006) https://doi.org/10.1038/nature04258
    - Single-trial method:    Richter et al (2015) https://doi.org/10.1016/j.neuroimage.2015.04.040
    """
    method = method.lower()
    if method in ['ppc','pairwise_phase_consistency']:  sync_func = ppc
    elif method in ['plv','phase_locking_value']:       sync_func = plv
    elif method in ['coh','coherence']:                 sync_func = coherence
    else:
        raise ValueError("Unsupported value set for <method>: '%s'" % method)

    return sync_func(data1, data2, axis=axis, return_phase=return_phase, single_trial=single_trial,
                     data_type=data_type, spec_method=spec_method, smp_rate=smp_rate,
                     time_axis=time_axis, taper_axis=taper_axis, keepdims=keepdims, **kwargs)


# =============================================================================
# Spike-Field Synchrony functions
# =============================================================================
def spike_field_coupling(spkdata, lfpdata, axis=0, method='PPC', return_phase=False,
                         time_axis=None, taper_axis=None, timepts=None,
                         data_type=None, spec_method='multitaper', smp_rate=None,
                         keepdims=True, **kwargs):
    """
    Computs measure of pairwise coupling between a pair of spike and continuous (eg LFP)
    raw or spectral (time-frequency) data, using given estimation method

    Parameters
    ----------
    spkdata : ndarray, shape=(...,n_obs,...), dtype=bool
        Single-channel binary spike trains (with 1's labelling spike times, 0's otherwise).

        For coherence: Can be given either as raw binary spike trains or as their spectral
        transform, but must have same data type (raw or spectral) and shape as `lfpdata`.

        For PLV/PPC: Shape is arbitrary, but MUST have same shape as `lfpdata`
        for raw `lfpdata`, and same dimensionality as `lfpdata` for spectral `lfpdata`.
        Thus, if `lfpdata` is spectral, must pre-pend singleton (length-1) axis to
        spkdata to match (eg using np.newaxis).

    lfpdata : ndarray, shape=(...,n_obs,...)
        Single-channel continuous (eg LFP) data.
        Can be given as raw LFPs or complex-valued time-frequency transform.

        Trial/observation axis is assumed to be axis 0 unless given in `axis`.
        For raw data, axis corresponding to time must be given in `time_axis`.

        Other than those constraints, data can have arbitrary shape, with analysis performed
        in mass-bivariate fashion independently along each axis other than observation `axis`
        (eg different frequencies, timepoints, conditions, etc.)

    axis : int, default: 0
        Axis corresponding to distinct observations/trials

    method : {'PPC','PLV','coherence'}, default: 'PPC'
        Spike-field coupling estimation method

        - 'PPC' : Pairwise Phase Consistency (see :func:`.spike_field_ppc`)
        - 'PLV' : Phase Locking Value (see :func:`.spike_field_plv`)
        - 'coherence' : coherence (see :func:`.spike_field_coherence`)

    return_phase : bool, default: False
        If False, only return measure of synchrony magnitude/strength between data1 and data2.
        If True, also returns mean LFP phase of spikes (or coherence phase) in additional output.

    time_axis : int
        Axis of data corresponding to time. Required input for phase-based methods (PLV/PPC).
        For coherence, only needed for `data_type` = 'raw' AND `spec_method` = 'multitaper'.

    taper_axis : int
        Axis of spectral data corresponding to tapers. Only needed for multitaper spectral data.

    timepts : array-like, shape=(n_timepts,), default: (0 - n_timepts-1)/smp_rate
        Time sampling vector for data. Default value starts at 0, with spacing = 1/smp_rate.
        For phase methods, should be in same time units as `width`/`spacing`/`lims` or `timewins`.

    data_type : {'raw','spectral'}, default: assume 'raw' if data is real; 'spectral' if complex
        What kind of data is input: 'raw' or 'spectral'?

    spec_method : {'wavelet','multitaper','bandfilter'}, default: 'wavelet'
        Method to use for (or already used for) spectral analysis.
        NOTE: Must be input for multitaper spectral data, so taper axis is handled appropriately.
        Otherwise, only used for *raw* data.

    smp_rate : scalar
        Sampling rate of data (only needed for *raw* data)

    keepdims : bool, default: True
        If True, retains reduced trial and/or taper axes as length-one axes in output.
        If False, removes reduced trial,taper axes from outputs.

    **kwargs :
        All other kwargs passed as-is to specific synchrony estimation function.

    Returns
    -------
    sync : ndarray, shape=(...,[1,]...)
        Magnitude/strength of spike-field coupling (coherence or PLV/PPC magnitude).

        For phase-based methods, time windows without any spikes are set = np.nan.

        If lfpdata is spectral, this has same shape as lfpdata, but time axis reduced
        to n_timewins and with `axis` (and taper_axis as well for multitaper) reduced
        to length 1 if `keepdims` is True or removed if `keepdims` is False.
        
        If lfpdata is raw, this has the same shape as described above for spectral data,
        but with a new frequency axis (and also a taper axis if `spec_method`='multitaper`)
        inserted immediately before `time_axis`.

    freqs : ndarray, shape=(n_freqs,)
        List of frequencies in `sync`. Only returned for raw data, [] otherwise.

    timepts : ndarray, shape=(n_timepts,)
        List of timepoints in `sync` (in s, referenced to start of data).
        Only returned for raw data, [] otherwise.

    n : ndarray, shape=(n_timepts,), dtype=int
        Number of spikes contributing to synchrony computations.
        Value only returned for phase-based measures (PLV/PPC), for coherence, returns None.

    phi : ndarray, shape=(...,[1,]...), optional
        Mean phase of LFPs at spike times (or phase of complex coherency) in radians.
        Same shape as `sync`.
        Optional: Only returned if return_phase is True.

    Examples
    --------
    sync,freqs,timepts,n = spike_field_coupling(spkdata,lfpdata,return_phase=False)

    sync,freqs,timepts,n,phi = spike_field_coupling(spkdata,lfpdata,return_phase=True)
    """
    method = method.lower()
    if method in ['ppc','pairwise_phase_consistency']:  sfc_func = spike_field_ppc
    elif method in ['plv','phase_locking_value']:       sfc_func = spike_field_plv
    elif method in ['coh','coherence']:                 sfc_func = spike_field_coherence
    else:
        raise ValueError("Unsupported value '%s' given for <method>. \
                         Should be 'PPC'|'PLV'|'coherence'" % method)

    return sfc_func(spkdata, lfpdata, axis=axis, time_axis=time_axis, taper_axis=taper_axis,
                    timepts=timepts, data_type=data_type, spec_method=spec_method,
                    smp_rate=smp_rate, return_phase=return_phase, keepdims=keepdims, **kwargs)


# =============================================================================
# Data simulation and testing functions
# =============================================================================
def simulate_multichannel_oscillation(n_chnls, *args, **kwargs):
    """
    Generate synthetic multichannel data with oscillations at given parameters.

    For each channel, generate multiple trials with constant oscillatory signal +
    random additive Gaussian noise. Parameters can be shared across channels or set
    independently for each channel.

    Parameters
    ----------
    n_chnls : int
        Number of channels to simulate

    *args :
    **kwargs :
        Rest of place and keyword arguments are passed to :func:`.simulate_oscillation`
        for each channel. Each argument can be given in one of two forms:

        (1) Scalar. A single value equivalent to the same argument to simulate_oscillation().
            That same value will be copied for all `n_chnls` channels.
        (2) Array-like, shape=(n_chnls,). List with different values for each channel,
            which will be iterated through.

        See :func:`.simulate_oscillation` for details on arguments.

        Exceptions: If a value is set for `seed`, it will be used to set the
        random number generator only ONCE at the start of this function, so that
        the generation of all channel signals follow a reproducible random sequence.

        Simulated data must have same shape for all channels, so all channels must
        have the same value set for `time_range`, `smp_rate`, and `n_trials`.

    Returns
    -------
    data : ndarray, shape=(n_timepts,n_trials,n_chnls)
        Simulated multichannel data
    """
    # Set a single random seed, so all multichannel data follows reproducible sequence
    seed = kwargs.pop('seed',None)
    if seed is not None:
        if not np.isscalar(seed) and (len(seed) > 1):
            seed = seed[0]
            print("Multiple values not permitted for <seed>. Using first one.")
        set_random_seed(seed)

    # Ensure all channels have same values for these parameters that determine data size
    for param in ['time_range','smp_rate','n_trials']:
        assert (param not in kwargs) or np.isscalar(kwargs[param]) or \
                np.allclose(np.diff(kwargs[param]), 0), \
            ValueError("All simulated channels must have same value for '%s'" % param)

    # Replicate any single-valued arguments (args and kwargs) into (n_chnls,) lists
    args = list(args)
    for i,value in enumerate(args):
        if np.isscalar(value) or (len(value) != n_chnls): args[i] = n_chnls * [value]

    for key,value in kwargs.items():
        if np.isscalar(value) or (len(value) != n_chnls): kwargs[key] = n_chnls * [value]

    # Simulate oscillatory data for each channel
    for chnl in range(n_chnls):
        # Extract the i-th value for each arg and kwarg
        chnl_args = [value[chnl] for value in args]
        chnl_kwargs = {key:value[chnl] for key,value in kwargs.items()}

        chnl_data = simulate_oscillation(*chnl_args, **chnl_kwargs)

        # HACK Create the array to hold all channel data after we know the shape
        if chnl == 0:
            n_timepts,n_trials = chnl_data.shape
            data = np.empty((n_timepts,n_trials,n_chnls))

        data[:,:,chnl] = chnl_data

    return data





# -*- coding: utf-8 -*-
""" Nonparametric bootstrap statistics """
import numpy as np

from spynal.utils import axis_index_slices, isarraylike, set_random_seed
from spynal.randstats.sampling import bootstraps
from spynal.randstats.utils import resamples_to_pvalue
from spynal.randstats.helpers import _str_to_one_sample_stat, _str_to_assoc_stat, \
                                     _str_to_two_sample_stat, _tail_to_compare, \
                                     _paired_sample_data_checks, _two_sample_data_checks


def one_sample_bootstrap_test(data, axis=0, mu=0, stat='t', tail='both', n_resamples=10000,
                              seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate 1-sample bootstrap test

    Parameters and returns are same as :func:`.one_sample_test`

    Computes stat on each bootstrap resample, and subtracts off stat computed on
    observed data to center resamples at 0 (mu) to estimate null distribution.
    p value is proportion of centered resampled values exceeding observed value.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch. 3.10
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data.ndim + axis

    # Convert string specifiers to callable functions
    stat_func    = _str_to_one_sample_stat(stat,axis)   # Statistic (includes default)
    compare_func = _tail_to_compare(tail)               # Tail-specific comparator

    ndim = data.ndim
    n = data.shape[axis]

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data, **kwargs)

    # Create generators with n_resamples-1 random bootstrap resamplings with replacement
    resamples = bootstraps(n, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)
            # Compute statistic on resampled data
            stat_resmp[stat_slices] = stat_func(data[data_slices], **kwargs)

        # Center each resampled statistic on their mean to approximate null disttribution
        # TODO Need to determine which of these is more apropriate (and align ~return_stats option)
        stat_resmp -= stat_obs
        # stat_resmp -= stat_resmp.mean(axis=axis,keepdims=True)

        # p value = proportion of bootstrap-resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs - mu, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            # Compute statistic on resampled data
            stat_resmp = stat_func(data[data_slices])

            # Subtract observed statistic from distribution of resampled statistics
            stat_resmp -= stat_obs

            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs - mu, stat_resmp, **kwargs)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isarraylike(stat_obs): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p


def paired_sample_bootstrap_test(data1, data2, axis=0, d=0, stat='t', tail='both',
                                 n_resamples=10000, seed=None, return_stats=False,
                                 keepdims=True, **kwargs):
    """
    Mass univariate paired-sample bootstrap test

    Parameters and returns are same as :func:`.paired_sample_test`

    Computes stat on each bootstrap resample, and subtracts off stat computed on
    observed data to center resamples at 0 (mu) to estimate null distribution.
    p value is proportion of centered resampled values exceeding observed value.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.6.1
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _paired_sample_data_checks(data1, data2)

    if isinstance(stat,str) and (stat.lower == 'meandiff'): stat = 'mean'

    return one_sample_bootstrap_test(data1 - data2, axis=axis, mu=d, stat=stat,
                                     tail=tail, n_resamples=n_resamples,
                                     return_stats=return_stats, seed=seed,
                                     keepdims=keepdims, **kwargs)


def paired_sample_association_bootstrap_test(data1, data2, axis=0, stat='r', tail='both',
                                             n_resamples=10000, seed=None, return_stats=False,
                                             keepdims=True, **kwargs):
    """
    Mass bivariate boostrap test of association (eg correlation) between two paired samples

    Parameters and returns are same as :func:`.paired_sample_association_test`

    Observations are bootstrap resampled in pairs, and stat is recomputed on each.
    Stat computed on observed data is subtracted off resamples, to center them at 0 and estimate
    null distribution. p value is proportion of centered resampled values exceeding observed value.
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _paired_sample_data_checks(data1, data2)

    # Convert string specifiers to callable functions
    stat_func    = _str_to_assoc_stat(stat,axis)    # Statistic (includes default)
    compare_func = _tail_to_compare(tail)           # Tail-specific comparator

    ndim = data1.ndim
    n = data1.shape[axis]

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data1,data2, **kwargs)

    # Create generators with n_resamples random bootstrap resamplings with replacement
    resamples = bootstraps(n, n_resamples-1, seed)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data1.shape)]
        stat_resmp = np.empty(stat_shape)
        for i_resmp,resample in enumerate(resamples):
            # Index into <axis> of data and stat, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)
            # Compute statistic on resampled data. Only need to permute one sample
            stat_resmp[stat_slices] = stat_func(data1[data_slices], data2[data_slices], **kwargs)

        # Center each resampled statistic on their mean to approximate null disttribution
        # TODO Need to determine which of these is more apropriate (and align ~return_stats option)
        stat_resmp -= stat_obs
        # stat_resmp -= stat_resmp.mean(axis=axis,keepdims=True)

        # p value = proportion of bootstrap-resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data1.shape)]
        count = np.ones(stat_shape)
        for resample in resamples:
            # Index into <axis> of data, with ':' for all other axes
            data_slices = axis_index_slices(axis, resample, ndim)
            # Compute statistic on resampled data1,data2
            stat_resmp  = stat_func(data1[data_slices], data2[data_slices], **kwargs)
            # Subtract observed statistic from resampled statistics
            stat_resmp -= stat_obs
            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs, stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isinstance(stat_obs,np.ndarray): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p


def two_sample_bootstrap_test(data1, data2, axis=0, stat='t', tail='both',
                              n_resamples=10000, seed=None, return_stats=False,
                              keepdims=True, **kwargs):
    """
    Mass univariate bootstrap two-sample test

    Parameters and returns are same as :func:`.two_sample_test`

    Computes stat on each pair of bootstrap resamples, and subtracts off stat computed
    on observed data to center resamples at 0 (mu) to estimate null distribution.
    p value is proportion of centered resampled values exceeding observed value.

    References
    ----------
    Manly (1997) "Randomization, Bootstrap and Monte Carlo Methods in Biology" ch.3.10, 6.3
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _two_sample_data_checks(data1, data2, axis)

    # Convert string specifiers to callable functions
    stat_func    = _str_to_two_sample_stat(stat,axis)   # Statistic (includes default)
    compare_func = _tail_to_compare(tail)               # Tail-specific comparator

    ndim = data1.ndim
    n1 = data1.shape[axis]
    n2 = data2.shape[axis]

    # Compute statistic of interest on actual observed data
    stat_obs = stat_func(data1, data2, **kwargs)

    # Create generators with n_resamples-1 random resamplings with replacement
    # of ints 0:n1-1 and 0:n2-1
    # Note: Seed random number generator only *once* before generating both random samples
    if seed is not None: set_random_seed(seed)
    resamples1 = bootstraps(n1,n_resamples-1)
    resamples2 = bootstraps(n2,n_resamples-1)

    if return_stats:
        # Compute statistic under <n_resamples> random resamplings
        stat_shape = [axlen if ax != axis else n_resamples-1 for ax,axlen in enumerate(data1.shape)]
        stat_resmp = np.empty(stat_shape)

        # Iterate thru <n_resamples> random resamplings, recomputing statistic on each
        for i_resmp,(resample1,resample2) in enumerate(zip(resamples1,resamples2)):
            # Index into <axis> of data and stat, with ':' for all other axes
            data1_slices = axis_index_slices(axis, resample1, ndim)
            data2_slices = axis_index_slices(axis, resample2, ndim)
            stat_slices = axis_index_slices(axis, [i_resmp], ndim)
            stat_resmp[stat_slices] = stat_func(data1[data1_slices], data2[data2_slices], **kwargs)

        # Center each resampled statistic on their mean to approximate null disttribution
        # TODO Need to determine which of these is more apropriate (and align ~return_stats option)
        stat_resmp -= stat_obs
        # stat_resmp -= stat_resmp.mean(axis=axis,keepdims=True)

        # p value = proportion of resampled test statistic values
        #  >= observed value (+ observed value itself)
        p = resamples_to_pvalue(stat_obs, stat_resmp, axis, compare_func)

    else:
        # Compute statistic under <n_resamples> random resamplings and
        # tally values exceeding given criterion (more extreme than observed)
        # Note: Init count to 1's to account for observed value itself
        stat_shape = [axlen if ax != axis else 1 for ax,axlen in enumerate(data1.shape)]
        count = np.ones(stat_shape)
        for resample1,resample2 in zip(resamples1,resamples2):
            # Index into <axis> of data and stat, with ':' for all other axes
            data1_slices = axis_index_slices(axis, resample1, ndim)
            data2_slices = axis_index_slices(axis, resample2, ndim)

            # Compute statistic on resampled data
            stat_resmp = stat_func(data1[data1_slices], data2[data2_slices], **kwargs)
            # Subtract observed statistic from distribution of resampled statistics
            stat_resmp -= stat_obs

            # Compare observed and resampled stats, and tally values passing criterion
            count += compare_func(stat_obs,stat_resmp)

        # p value = proportion of resampled test statistic values passing criterion
        p = count / n_resamples

    # For vector-valued data, extract value from scalar array -> float for output
    if p.size == 1:
        p = p.item()
        if return_stats and isarraylike(stat_obs): stat_obs = stat_obs.item()
    elif not keepdims:
        p = p.squeeze(axis=axis)
        if return_stats: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return p, stat_obs, stat_resmp
    else:               return p

